"""
Functions about validation.
"""
from __future__ import annotations
import re
import pytz
import datetime as dt
from typing import Optional, Union, TYPE_CHECKING

import pycountry
import numpy as np
import pandas as pd
from pandas import DataFrame

from . import constants as cs
from . import helpers as hp

if TYPE_CHECKING:
    from .feed import Feed


TIME_PATTERN1 = re.compile(r"^\d\d:\d\d:\d\d$")
TIME_PATTERN2 = re.compile(r"^\d:\d\d:\d\d$")
DATE_FORMAT = "%Y%m%d"
TIMEZONES = set(pytz.all_timezones)
# ISO639-1 language codes, both lower and upper case
LANGS = set([lang.alpha_2 for lang in pycountry.languages if hasattr(lang, "alpha_2")])
LANGS |= set(x.upper() for x in LANGS)
CURRENCIES = set([c.alpha_3 for c in pycountry.currencies if hasattr(c, "alpha_3")])
URL_PATTERN = re.compile(
    r"^(?:http)s?://"  # http:// or https://
    r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # domain...
    r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
    r"(?::\d+)?"  # optional port
    r"(?:/?|[/?]\S+)$",
    re.IGNORECASE,
)
EMAIL_PATTERN = re.compile(r"[^@]+@[^@]+\.[^@]+")
COLOR_PATTERN = re.compile(r"(?:[0-9a-fA-F]{2}){3}$")


def valid_str(x: str) -> bool:
    """
    Return ``True`` if ``x`` is a non-blank string;
    otherwise return ``False``.
    """
    if isinstance(x, str) and x.strip():
        return True
    else:
        return False


def valid_time(x: str) -> bool:
    """
    Return ``True`` if ``x`` is a valid H:MM:SS or HH:MM:SS time;
    otherwise return ``False``.
    """
    if isinstance(x, str) and (
        re.match(TIME_PATTERN1, x) or re.match(TIME_PATTERN2, x)
    ):
        return True
    else:
        return False


def valid_date(x: str) -> bool:
    """
    Retrun ``True`` if ``x`` is a valid YYYYMMDD date;
    otherwise return ``False``.
    """
    try:
        if x != dt.datetime.strptime(x, DATE_FORMAT).strftime(DATE_FORMAT):
            raise ValueError
        return True
    except ValueError:
        return False


def valid_timezone(x: str) -> bool:
    """
    Retrun ``True`` if ``x`` is a valid human-readable timezone string,
    e.g. 'Africa/Abidjan'; otherwise return ``False``.
    """
    return x in TIMEZONES


def valid_lang(x: str) -> bool:
    """
    Return ``True`` if ``x`` is a valid two-letter ISO 639 language
    code, e.g. 'aa'; otherwise return ``False``.
    """
    return x in LANGS


def valid_currency(x: str) -> bool:
    """
    Return ``True`` if ``x`` is a valid three-letter ISO 4217 currency
    code, e.g. 'AED'; otherwise return ``False``.
    """
    return x in CURRENCIES


def valid_url(x: str) -> bool:
    """
    Return ``True`` if ``x`` is a valid URL; otherwise return ``False``.
    """
    if isinstance(x, str) and re.match(URL_PATTERN, x):
        return True
    else:
        return False


def valid_email(x: str) -> bool:
    """
    Return ``True`` if ``x`` is a valid email address; otherwise return
    ``False``.
    """
    if isinstance(x, str) and re.match(EMAIL_PATTERN, x):
        return True
    else:
        return False


def valid_color(x: str) -> bool:
    """
    Return ``True`` if ``x`` a valid hexadecimal color string without
    the leading hash; otherwise return ``False``.
    """
    if isinstance(x, str) and re.match(COLOR_PATTERN, x):
        return True
    else:
        return False


def check_for_required_columns(problems: list, table: str, df: DataFrame) -> list:
    """
    Check that the given GTFS table has the required columns.

    Parameters
    ----------
    problems : list
        A four-tuple containing

        1. A problem type (string) equal to ``'error'`` or ``'warning'``;
           ``'error'`` means the GTFS is violated;
           ``'warning'`` means there is a problem but it is not a
           GTFS violation
        2. A message (string) that describes the problem
        3. A GTFS table name, e.g. ``'routes'``, in which the problem
           occurs
        4. A list of rows (integers) of the table's DataFrame where the
           problem occurs

    table : string
        Name of a GTFS table
    df : DataFrame
        The GTFS table corresponding to ``table``

    Returns
    -------
    list
        The ``problems`` list extended as follows.
        Check that the DataFrame contains the colums required by GTFS
        and append to the problems list one error for each column
        missing.

    """
    r = cs.GTFS_REF
    req_columns = r.loc[(r["table"] == table) & r["column_required"], "column"].values
    for col in req_columns:
        if col not in df.columns:
            problems.append(["error", f"Missing column {col}", table, []])

    return problems


def check_for_invalid_columns(problems: list, table: str, df: DataFrame) -> list:
    """
    Check for invalid columns in the given GTFS DataFrame.

    Parameters
    ----------
    problems : list
        A four-tuple containing

        1. A problem type (string) equal to ``'error'`` or ``'warning'``;
           ``'error'`` means the GTFS is violated;
           ``'warning'`` means there is a problem but it is not a
           GTFS violation
        2. A message (string) that describes the problem
        3. A GTFS table name, e.g. ``'routes'``, in which the problem
           occurs
        4. A list of rows (integers) of the table's DataFrame where the
           problem occurs

    table : string
        Name of a GTFS table
    df : DataFrame
        The GTFS table corresponding to ``table``

    Returns
    -------
    list
        The ``problems`` list extended as follows.
        Check whether the DataFrame contains extra columns not in the
        GTFS and append to the problems list one warning for each extra
        column.

    """
    r = cs.GTFS_REF
    valid_columns = r.loc[r["table"] == table, "column"].values
    for col in df.columns:
        if col not in valid_columns:
            problems.append(["warning", f"Unrecognized column {col}", table, []])

    return problems


def check_table(
    problems: list,
    table: str,
    df: DataFrame,
    condition,
    message: str,
    type_: str = "error",
) -> list:
    """
    Check the given GTFS table for the given problem condition.

    Parameters
    ----------
    problems : list
        A four-tuple containing

        1. A problem type (string) equal to ``'error'`` or ``'warning'``;
           ``'error'`` means the GTFS is violated;
           ``'warning'`` means there is a problem but it is not a
           GTFS violation
        2. A message (string) that describes the problem
        3. A GTFS table name, e.g. ``'routes'``, in which the problem
           occurs
        4. A list of rows (integers) of the table's DataFrame where the
           problem occurs

    table : string
        Name of a GTFS table
    df : DataFrame
        The GTFS table corresponding to ``table``
    condition : boolean expression
        One involving ``df``, e.g.`df['route_id'].map(is_valid_str)``
    message : string
        Problem message, e.g. ``'Invalid route_id'``
    type_ : string
        ``'error'`` or ``'warning'`` indicating the type of problem
        encountered

    Returns
    -------
    list
        The ``problems`` list extended as follows.
        Record the indices of ``df`` that statisfy the condition.
        If the list of indices is nonempty, append to the
        problems the item ``[type_, message, table, indices]``;
        otherwise do not append anything.

    """
    indices = df.loc[condition].index.tolist()
    if indices:
        problems.append([type_, message, table, indices])

    return problems


def check_column(
    problems: list,
    table: str,
    df: DataFrame,
    column: str,
    checker,
    message: Optional[str] = None,
    type_: str = "error",
    *,
    column_required: bool = True,
) -> list:
    """
    Check the given column of the given GTFS with the given problem
    checker.

    Parameters
    ----------
    problems : list
        A four-tuple containing

        1. A problem type (string) equal to ``'error'`` or ``'warning'``;
           ``'error'`` means the GTFS is violated;
           ``'warning'`` means there is a problem but it is not a
           GTFS violation
        2. A message (string) that describes the problem
        3. A GTFS table name, e.g. ``'routes'``, in which the problem
           occurs
        4. A list of rows (integers) of the table's DataFrame where the
           problem occurs

    table : string
        Name of a GTFS table
    df : DataFrame
        The GTFS table corresponding to ``table``
    column : string
        A column of ``df``
    column_required : boolean
        ``True`` if and only if ``column`` is required
        (and not optional) by the GTFS
    checker : boolean valued unary function
        Returns ``True`` if and only if no problem is encountered
    message : string (optional)
        Problem message, e.g. 'Invalid route_id'.
        Defaults to 'Invalid ``column``; maybe has extra space characters'
    type_ : string
        ``'error'`` or ``'warning'`` indicating the type of problem
        encountered

    Returns
    -------
    list
        The ``problems`` list extended as follows.
        Apply the checker to the column entries and record the indices
        of ``df`` where the checker returns ``False``.
        If the list of indices of is nonempty, append to the problems the
        item ``[type_, problem, table, indices]``; otherwise do not
        append anything.

        If not ``column_required``, then NaN entries will be ignored
        before applying the checker.

    """
    f = df.copy()
    if not column_required:
        if column not in f.columns:
            f[column] = np.nan
        f = f.dropna(subset=[column])

    cond = ~f[column].map(checker)

    if not message:
        message = f"Invalid {column}; maybe has extra space characters"

    problems = check_table(problems, table, f, cond, message, type_)

    return problems


def check_column_id(
    problems: list,
    table: str,
    df: DataFrame,
    column: str,
    *,
    column_required: bool = True,
) -> list:
    """
    A specialization of :func:`check_column`.

    Parameters
    ----------
    problems : list
        A four-tuple containing

        1. A problem type (string) equal to ``'error'`` or ``'warning'``;
           ``'error'`` means the GTFS is violated;
           ``'warning'`` means there is a problem but it is not a
           GTFS violation
        2. A message (string) that describes the problem
        3. A GTFS table name, e.g. ``'routes'``, in which the problem
           occurs
        4. A list of rows (integers) of the table's DataFrame where the
           problem occurs

    table : string
        Name of a GTFS table
    df : DataFrame
        The GTFS table corresponding to ``table``
    column : string
        A column of ``df``
    column_required : boolean
        ``True`` if and only if ``column`` is required
        (and not optional) by the GTFS

    Returns
    -------
    list
        The ``problems`` list extended as follows.
        Record the indices of ``df`` where the given column has
        duplicated entry or an invalid strings.
        If the list of indices is nonempty, append to the problems the
        item ``[type_, problem, table, indices]``; otherwise do not
        append anything.

        If not ``column_required``, then NaN entries will be ignored
        in the checking.

    """
    f = df.copy()
    if not column_required:
        if column not in f.columns:
            f[column] = np.nan
        f = f.dropna(subset=[column])

    cond = ~f[column].map(valid_str)
    problems = check_table(
        problems,
        table,
        f,
        cond,
        f"Invalid {column}; maybe has extra space characters",
    )

    cond = f[column].duplicated()
    problems = check_table(problems, table, f, cond, f"Repeated {column}")

    return problems


def check_column_linked_id(
    problems: list,
    table: str,
    df: DataFrame,
    column: str,
    target_df: DataFrame,
    target_column: Optional[str] = None,
    *,
    column_required: bool = True,
) -> list:
    """
    A modified version of :func:`check_column_id`.

    Parameters
    ----------
    problems : list
        A four-tuple containing

        1. A problem type (string) equal to ``'error'`` or ``'warning'``;
           ``'error'`` means the GTFS is violated;
           ``'warning'`` means there is a problem but it is not a
           GTFS violation
        2. A message (string) that describes the problem
        3. A GTFS table name, e.g. ``'routes'``, in which the problem
           occurs
        4. A list of rows (integers) of the table's DataFrame where the
           problem occurs

    table : string
        Name of a GTFS table
    df : DataFrame
        The GTFS table corresponding to ``table``
    column : string
        A column of ``df``
    column_required : boolean
        ``True`` if and only if ``column`` is required
        (and not optional) by the GTFS
    target_df : DataFrame
        A GTFS table
    target_column : string
        A column of ``target_df``; defaults to ``column_name``

    Returns
    -------
    list
        The ``problems`` list extended as follows.
        Record indices of ``df`` where the following condition is
        violated: ``column`` contain IDs that are valid strings and are
        present in ``target_df`` under the ``target_column`` name.
        If the list of indices is nonempty, append to the problems the
        item ``[type_, problem, table, indices]``; otherwise do not
        append anything.

        If not ``column_required``, then NaN entries will be ignored
        in the checking.

    """
    if target_column is None:
        target_column = column

    f = df.copy()

    if target_df is None:
        g = pd.DataFrame()
        g[target_column] = np.nan
    else:
        g = target_df.copy()

    if target_column not in g.columns:
        g[target_column] = np.nan

    if not column_required:
        if column not in f.columns:
            f[column] = np.nan
        f = f.dropna(subset=[column])
        g = g.dropna(subset=[target_column])

    cond = ~f[column].isin(g[target_column])
    problems = check_table(problems, table, f, cond, f"Undefined {column}")

    return problems


def format_problems(problems: list, *, as_df: bool = False) -> Union[list, DataFrame]:
    """
    Format the given problems list as a DataFrame.

    Parameters
    ----------
    problems : list
        A four-tuple containing

        1. A problem type (string) equal to ``'error'`` or ``'warning'``;
           ``'error'`` means the GTFS is violated;
           ``'warning'`` means there is a problem but it is not a
           GTFS violation
        2. A message (string) that describes the problem
        3. A GTFS table name, e.g. ``'routes'``, in which the problem
           occurs
        4. A list of rows (integers) of the table's DataFrame where the
           problem occurs

    as_df : boolean

    Returns
    -------
    list or DataFrame
        Return ``problems`` if not ``as_df``; otherwise return a
        DataFrame with the problems as rows and the columns
        ``['type', 'message', 'table', 'rows']``.

    """
    if as_df:
        problems = pd.DataFrame(
            problems, columns=["type", "message", "table", "rows"]
        ).sort_values(["type", "table"])
    return problems


def check_agency(
    feed: "Feed", *, as_df: bool = False, include_warnings: bool = False
) -> list:
    """
    Check that ``feed.agency`` follows the GTFS.
    Return a list of problems of the form described in
    :func:`check_table`;
    the list will be empty if no problems are found.
    """
    table = "agency"
    problems = []

    # Preliminary checks
    if feed.agency is None:
        problems.append(["error", "Missing table", table, []])
    else:
        f = feed.agency.copy()
        problems = check_for_required_columns(problems, table, f)
    if problems:
        return format_problems(problems, as_df=as_df)

    if include_warnings:
        problems = check_for_invalid_columns(problems, table, f)

    # Check service_id
    problems = check_column_id(problems, table, f, "agency_id", column_required=False)

    # Check agency_name
    problems = check_column(problems, table, f, "agency_name", valid_str)

    # Check agency_url
    problems = check_column(problems, table, f, "agency_url", valid_url)

    # Check agency_timezone
    problems = check_column(problems, table, f, "agency_timezone", valid_timezone)

    # Check agency_fare_url
    problems = check_column(
        problems, table, f, "agency_fare_url", valid_url, column_required=False
    )

    # Check agency_lang
    problems = check_column(
        problems, table, f, "agency_lang", valid_lang, column_required=False
    )

    # Check agency_phone
    problems = check_column(
        problems, table, f, "agency_phone", valid_str, column_required=False
    )

    # Check agency_email
    problems = check_column(
        problems, table, f, "agency_email", valid_email, column_required=False
    )

    return format_problems(problems, as_df=as_df)


def check_calendar(
    feed: "Feed", *, as_df: bool = False, include_warnings: bool = False
) -> list:
    """
    Analog of :func:`check_agency` for ``feed.calendar``.
    """
    table = "calendar"
    problems = []

    # Preliminary checks
    if feed.calendar is None:
        return problems

    f = feed.calendar.copy()
    problems = check_for_required_columns(problems, table, f)
    if problems:
        return format_problems(problems, as_df=as_df)

    if include_warnings:
        problems = check_for_invalid_columns(problems, table, f)

    # Check service_id
    problems = check_column_id(problems, table, f, "service_id")

    # Check weekday columns
    v = lambda x: x in range(2)
    for col in [
        "monday",
        "tuesday",
        "wednesday",
        "thursday",
        "friday",
        "saturday",
        "sunday",
    ]:
        problems = check_column(problems, table, f, col, v)

    # Check start_date and end_date
    for col in ["start_date", "end_date"]:
        problems = check_column(problems, table, f, col, valid_date)

    if include_warnings:
        # Check if feed has expired
        d = f["end_date"].max()
        if feed.calendar_dates is not None and not feed.calendar_dates.empty:
            table += "/calendar_dates"
            d = max(d, feed.calendar_dates["date"].max())
        if d < dt.datetime.today().strftime(DATE_FORMAT):
            problems.append(["warning", "Feed expired", table, []])

    return format_problems(problems, as_df=as_df)


def check_calendar_dates(
    feed: "Feed", *, as_df: bool = False, include_warnings: bool = False
) -> list:
    """
    Analog of :func:`check_agency` for ``feed.calendar_dates``.
    """
    table = "calendar_dates"
    problems = []

    # Preliminary checks
    if feed.calendar_dates is None:
        return problems

    f = feed.calendar_dates.copy()
    problems = check_for_required_columns(problems, table, f)
    if problems:
        return format_problems(problems, as_df=as_df)

    if include_warnings:
        problems = check_for_invalid_columns(problems, table, f)

    # Check service_id
    problems = check_column(problems, table, f, "service_id", valid_str)

    # Check date
    problems = check_column(problems, table, f, "date", valid_date)

    # No duplicate (service_id, date) pairs allowed
    cond = f[["service_id", "date"]].duplicated()
    problems = check_table(problems, table, f, cond, "Repeated pair (service_id, date)")

    # Check exception_type
    v = lambda x: x in [1, 2]
    problems = check_column(problems, table, f, "exception_type", v)

    return format_problems(problems, as_df=as_df)


def check_fare_attributes(
    feed: "Feed", *, as_df: bool = False, include_warnings: bool = False
) -> list:
    """
    Analog of :func:`check_agency` for ``feed.calendar_dates``.
    """
    table = "fare_attributes"
    problems = []

    # Preliminary checks
    if feed.fare_attributes is None:
        return problems

    f = feed.fare_attributes.copy()
    problems = check_for_required_columns(problems, table, f)
    if problems:
        return format_problems(problems, as_df=as_df)

    if include_warnings:
        problems = check_for_invalid_columns(problems, table, f)

    # Check fare_id
    problems = check_column_id(problems, table, f, "fare_id")

    # Check currency_type
    problems = check_column(problems, table, f, "currency_type", valid_currency)

    # Check payment_method
    v = lambda x: x in range(2)
    problems = check_column(problems, table, f, "payment_method", v)

    # Check transfers
    v = lambda x: pd.isna(x) or x in range(3)
    problems = check_column(problems, table, f, "transfers", v)

    # Check transfer_duration
    v = lambda x: x >= 0
    problems = check_column(
        problems, table, f, "transfer_duration", v, column_required=False
    )

    return format_problems(problems, as_df=as_df)


def check_fare_rules(
    feed: "Feed", *, as_df: bool = False, include_warnings: bool = False
) -> list:
    """
    Analog of :func:`check_agency` for ``feed.calendar_dates``.
    """
    table = "fare_rules"
    problems = []

    # Preliminary checks
    if feed.fare_rules is None:
        return problems

    f = feed.fare_rules.copy()
    problems = check_for_required_columns(problems, table, f)
    if problems:
        return format_problems(problems, as_df=as_df)

    if include_warnings:
        problems = check_for_invalid_columns(problems, table, f)

    # Check fare_id
    problems = check_column_linked_id(
        problems, table, f, "fare_id", feed.fare_attributes
    )

    # Check route_id
    problems = check_column_linked_id(
        problems, table, f, "route_id", feed.routes, column_required=False
    )

    # Check origin_id, destination_id, contains_id
    for col in ["origin_id", "destination_id", "contains_id"]:
        problems = check_column_linked_id(
            problems,
            table,
            f,
            col,
            feed.stops,
            "zone_id",
            column_required=False,
        )

    return format_problems(problems, as_df=as_df)


def check_feed_info(
    feed: "Feed", *, as_df: bool = False, include_warnings: bool = False
) -> list:
    """
    Analog of :func:`check_agency` for ``feed.feed_info``.
    """
    table = "feed_info"
    problems = []

    # Preliminary checks
    if feed.feed_info is None:
        return problems

    f = feed.feed_info.copy()
    problems = check_for_required_columns(problems, table, f)
    if problems:
        return format_problems(problems, as_df=as_df)

    if include_warnings:
        problems = check_for_invalid_columns(problems, table, f)

    # Check feed_publisher_name
    problems = check_column(problems, table, f, "feed_publisher_name", valid_str)

    # Check feed_publisher_url
    problems = check_column(problems, table, f, "feed_publisher_url", valid_url)

    # Check feed_lang
    problems = check_column(problems, table, f, "feed_lang", valid_lang)

    # Check feed_start_date and feed_end_date
    cols = ["feed_start_date", "feed_end_date"]
    for col in cols:
        problems = check_column(
            problems, table, f, col, valid_date, column_required=False
        )

    if set(cols) <= set(f.columns):
        d1, d2 = f.loc[0, ["feed_start_date", "feed_end_date"]].values
        if pd.notna(d1) and pd.notna(d2) and d1 > d1:
            problems.append(
                [
                    "error",
                    "feed_start_date later than feed_end_date",
                    table,
                    [0],
                ]
            )

    # Check feed_version
    problems = check_column(
        problems, table, f, "feed_version", valid_str, column_required=False
    )

    return format_problems(problems, as_df=as_df)


def check_frequencies(
    feed: "Feed", *, as_df: bool = False, include_warnings: bool = False
) -> list:
    """
    Analog of :func:`check_agency` for ``feed.frequencies``.
    """
    table = "frequencies"
    problems = []

    # Preliminary checks
    if feed.frequencies is None:
        return problems

    f = feed.frequencies.copy()
    problems = check_for_required_columns(problems, table, f)
    if problems:
        return format_problems(problems, as_df=as_df)

    if include_warnings:
        problems = check_for_invalid_columns(problems, table, f)

    # Check trip_id
    problems = check_column_linked_id(problems, table, f, "trip_id", feed.trips)

    # Check start_time and end_time
    time_cols = ["start_time", "end_time"]
    for col in time_cols:
        problems = check_column(problems, table, f, col, valid_time)

    for col in time_cols:
        f[col] = f[col].map(hp.timestr_to_seconds)

    # Start_time should be earlier than end_time
    cond = f["start_time"] >= f["end_time"]
    problems = check_table(
        problems, table, f, cond, "start_time not earlier than end_time"
    )

    # Headway periods should not overlap
    f = f.sort_values(["trip_id", "start_time"])
    for __, group in f.groupby("trip_id"):
        a = group["start_time"].values
        b = group["end_time"].values
        indices = np.flatnonzero(a[1:] < b[:-1]).tolist()
        if indices:
            problems.append(
                [
                    "error",
                    "Headway periods for the same trip overlap",
                    table,
                    indices,
                ]
            )

    # Check headway_secs
    v = lambda x: x >= 0
    problems = check_column(problems, table, f, "headway_secs", v)

    # Check exact_times
    v = lambda x: x in range(2)
    problems = check_column(problems, table, f, "exact_times", v, column_required=False)

    return format_problems(problems, as_df=as_df)


def check_routes(
    feed: "Feed", *, as_df: bool = False, include_warnings: bool = False
) -> list:
    """
    Analog of :func:`check_agency` for ``feed.routes``.
    """
    table = "routes"
    problems = []

    # Preliminary checks
    if feed.routes is None:
        problems.append(["error", "Missing table", table, []])
    else:
        f = feed.routes.copy()
        problems = check_for_required_columns(problems, table, f)
    if problems:
        return format_problems(problems, as_df=as_df)

    if include_warnings:
        problems = check_for_invalid_columns(problems, table, f)

    # Check route_id
    problems = check_column_id(problems, table, f, "route_id")

    # Check agency_id
    if "agency_id" in f:
        if feed.agency is None:
            problems.append(
                [
                    "error",
                    "agency_id column present in routes agency table missing",
                    table,
                    [],
                ]
            )
        elif "agency_id" not in feed.agency.columns:
            problems.append(
                [
                    "error",
                    "agency_id column present in routes but not in agency",
                    table,
                    [],
                ]
            )
        else:
            g = f.dropna(subset=["agency_id"])
            cond = ~g["agency_id"].isin(feed.agency["agency_id"])
            problems = check_table(problems, table, g, cond, "Undefined agency_id")

    # Check route_short_name and route_long_name
    for column in ["route_short_name", "route_long_name"]:
        problems = check_column(
            problems, table, f, column, valid_str, column_required=False
        )

    cond = ~(f["route_short_name"].notna() | f["route_long_name"].notna())
    problems = check_table(
        problems,
        table,
        f,
        cond,
        "route_short_name and route_long_name both empty",
    )

    # Check route_type
    v = lambda x: x in range(8)
    problems = check_column(problems, table, f, "route_type", v)

    # Check route_url
    problems = check_column(
        problems, table, f, "route_url", valid_url, column_required=False
    )

    # Check route_color and route_text_color
    for col in ["route_color", "route_text_color"]:
        problems = check_column(
            problems, table, f, col, valid_color, column_required=False
        )

    if include_warnings:
        # Check for duplicated (route_short_name, route_long_name) pairs
        cond = f[["route_short_name", "route_long_name"]].duplicated()
        problems = check_table(
            problems,
            table,
            f,
            cond,
            "Repeated pair (route_short_name, route_long_name)",
            "warning",
        )

        # Check for routes without trips
        s = feed.trips["route_id"]
        cond = ~f["route_id"].isin(s)
        problems = check_table(
            problems, table, f, cond, "Route has no trips", "warning"
        )

    return format_problems(problems, as_df=as_df)


def check_shapes(
    feed: "Feed", *, as_df: bool = False, include_warnings: bool = False
) -> list:
    """
    Analog of :func:`check_agency` for ``feed.shapes``.
    """
    table = "shapes"
    problems = []

    # Preliminary checks
    if feed.shapes is None:
        return problems

    f = feed.shapes.copy()
    problems = check_for_required_columns(problems, table, f)
    if problems:
        return format_problems(problems, as_df=as_df)
    f.sort_values(["shape_id", "shape_pt_sequence"], inplace=True)

    if include_warnings:
        problems = check_for_invalid_columns(problems, table, f)

    # Check shape_id
    problems = check_column(problems, table, f, "shape_id", valid_str)

    # Check shape_pt_lon and shape_pt_lat
    for column, bound in [("shape_pt_lon", 180), ("shape_pt_lat", 90)]:
        v = lambda x: pd.notna(x) and -bound <= x <= bound
        cond = ~f[column].map(v)
        problems = check_table(
            problems,
            table,
            f,
            cond,
            f"{column} out of bounds {[-bound, bound]}",
        )

    # Check for duplicated (shape_id, shape_pt_sequence) pairs
    cond = f[["shape_id", "shape_pt_sequence"]].duplicated()
    problems = check_table(
        problems, table, f, cond, "Repeated pair (shape_id, shape_pt_sequence)"
    )

    # Check if shape_dist_traveled does decreases on a trip
    if "shape_dist_traveled" in f.columns:
        g = f.dropna(subset=["shape_dist_traveled"])
        indices = []
        prev_sid = None
        prev_index = None
        prev_dist = -1
        cols = ["shape_id", "shape_dist_traveled"]
        for i, sid, dist in g[cols].itertuples():
            if sid == prev_sid and dist < prev_dist:
                indices.append(prev_index)

            prev_sid = sid
            prev_index = i
            prev_dist = dist

        if indices:
            problems.append(
                [
                    "error",
                    "shape_dist_traveled decreases on a trip",
                    table,
                    indices,
                ]
            )

    return format_problems(problems, as_df=as_df)


def check_stops(
    feed: "Feed", *, as_df: bool = False, include_warnings: bool = False
) -> list:
    """
    Analog of :func:`check_agency` for ``feed.stops``.
    """
    table = "stops"
    problems = []

    # Preliminary checks
    if feed.stops is None:
        problems.append(["error", "Missing table", table, []])
    else:
        f = feed.stops.copy()
        problems = check_for_required_columns(problems, table, f)
    if problems:
        return format_problems(problems, as_df=as_df)

    if include_warnings:
        problems = check_for_invalid_columns(problems, table, f)

    # Check stop_id
    problems = check_column_id(problems, table, f, "stop_id")

    # Check stop_code, stop_desc, zone_id, parent_station
    for column in ["stop_code", "stop_desc", "zone_id", "parent_station"]:
        problems = check_column(
            problems, table, f, column, valid_str, column_required=False
        )

    # Check stop_name
    problems = check_column(problems, table, f, "stop_name", valid_str)

    # Check stop_lon and stop_lat
    if "location_type" in f.columns:
        requires_location = f.location_type.isin([0, 1, 2])
    else:
        requires_location = True
    for column, bound in [("stop_lon", 180), ("stop_lat", 90)]:
        v = lambda x: pd.notna(x) and -bound <= x <= bound
        cond = requires_location & ~f[column].map(v)
        problems = check_table(
            problems,
            table,
            f,
            cond,
            f"{column} out of bounds {[-bound, bound]}",
        )

    # Check stop_url
    problems = check_column(
        problems, table, f, "stop_url", valid_url, column_required=False
    )

    # Check location_type
    v = lambda x: x in range(5)
    problems = check_column(
        problems, table, f, "location_type", v, column_required=False
    )

    # Check stop_timezone
    problems = check_column(
        problems,
        table,
        f,
        "stop_timezone",
        valid_timezone,
        column_required=False,
    )

    # Check wheelchair_boarding
    v = lambda x: x in range(3)
    problems = check_column(
        problems, table, f, "wheelchair_boarding", v, column_required=False
    )

    # Check further location_type and parent_station
    if "parent_station" in f.columns:
        if "location_type" not in f.columns:
            problems.append(
                [
                    "error",
                    "parent_station column present but location_type column missing",
                    table,
                    [],
                ]
            )
        else:
            # Parent stations must be well-defined
            S = set(f.stop_id) | {np.nan}
            v = lambda x: x in S
            problems = check_column(
                problems,
                table,
                f,
                "parent_station",
                v,
                "A parent station must be well-defined",
                column_required=False,
            )

            # Stations must have location type 1
            station_ids = f.loc[f.parent_station.notna(), "parent_station"]
            cond = f.stop_id.isin(station_ids) & (f.location_type != 1)
            problems = check_table(
                problems, table, f, cond, "A station must have location_type 1"
            )

            # Stations must not lie in stations
            cond = (f.location_type == 1) & f.parent_station.notna()
            problems = check_table(
                problems,
                table,
                f,
                cond,
                "A station must not lie in another station",
            )

            # Entrances (type 2), generic nodes (type 3) and boarding areas (type 4)
            # need to be part of a parent
            cond = f.location_type.isin([2, 3, 4]) & f.parent_station.isna()
            problems = check_table(
                problems,
                table,
                f,
                cond,
                "Entrances, nodes, and boarding areas must be part of a parent station",
            )

    if include_warnings:
        # Check for stops of location type 0 or NaN without stop times
        ids = []
        if feed.stop_times is not None:
            ids = feed.stop_times.stop_id.unique()

        cond = ~feed.stops.stop_id.isin(ids)
        if "location_type" in feed.stops.columns:
            cond &= f.location_type.isin([0, np.nan])

        problems = check_table(
            problems, table, f, cond, "Stop has no stop times", "warning"
        )

    return format_problems(problems, as_df=as_df)


def check_stop_times(
    feed: "Feed", *, as_df: bool = False, include_warnings: bool = False
) -> list:
    """
    Analog of :func:`check_agency` for ``feed.stop_times``.
    """
    table = "stop_times"
    problems = []

    # Preliminary checks
    if feed.stop_times is None:
        problems.append(["error", "Missing table", table, []])
    else:
        f = feed.stop_times.copy().sort_values(["trip_id", "stop_sequence"])
        problems = check_for_required_columns(problems, table, f)
    if problems:
        return format_problems(problems, as_df=as_df)

    if include_warnings:
        problems = check_for_invalid_columns(problems, table, f)

    # Check trip_id
    problems = check_column_linked_id(problems, table, f, "trip_id", feed.trips)

    # Check arrival_time and departure_time
    v = lambda x: pd.isna(x) or valid_time(x)
    for col in ["arrival_time", "departure_time"]:
        problems = check_column(problems, table, f, col, v)

    # Check that arrival and departure times exist for the first and last
    # stop of each trip and for each timepoint.
    # For feeds with many trips, iterating through the stop time rows is
    # faster than uisg groupby.
    if "timepoint" not in f.columns:
        f["timepoint"] = np.nan  # This will not mess up later timepoint check

    indices = []
    prev_tid = None
    prev_index = None
    prev_atime = 1
    prev_dtime = 1
    for i, tid, atime, dtime, tp in f[
        ["trip_id", "arrival_time", "departure_time", "timepoint"]
    ].itertuples():
        if tid != prev_tid:
            # Check last stop of previous trip
            if pd.isna(prev_atime) or pd.isna(prev_dtime):
                indices.append(prev_index)
            # Check first stop of current trip
            if pd.isna(atime) or pd.isna(dtime):
                indices.append(i)
        elif tp == 1 and (pd.isna(atime) or pd.isna(dtime)):
            # Failure at timepoint
            indices.append(i)

        prev_tid = tid
        prev_index = i
        prev_atime = atime
        prev_dtime = dtime

    if pd.isna(prev_atime) or pd.isna(prev_dtime):
        indices.append(prev_index)

    if indices:
        problems.append(
            [
                "error",
                "First/last/time point arrival/departure time missing",
                table,
                indices,
            ]
        )

    # Check stop_id
    problems = check_column_linked_id(problems, table, f, "stop_id", feed.stops)

    # Check for duplicated (trip_id, stop_sequence) pairs
    cond = f[["trip_id", "stop_sequence"]].dropna().duplicated()
    problems = check_table(
        problems, table, f, cond, "Repeated pair (trip_id, stop_sequence)"
    )

    # Check stop_headsign
    problems = check_column(
        problems, table, f, "stop_headsign", valid_str, column_required=False
    )

    # Check pickup_type and drop_off_type
    for col in ["pickup_type", "drop_off_type"]:
        v = lambda x: x in range(4)
        problems = check_column(problems, table, f, col, v, column_required=False)

    # Check if shape_dist_traveled decreases on a trip
    if "shape_dist_traveled" in f.columns:
        g = f.dropna(subset=["shape_dist_traveled"])
        indices = []
        prev_tid = None
        prev_dist = -1
        for i, tid, dist in g[["trip_id", "shape_dist_traveled"]].itertuples():
            if tid == prev_tid and dist < prev_dist:
                indices.append(i)

            prev_tid = tid
            prev_dist = dist

        if indices:
            problems.append(
                [
                    "error",
                    "shape_dist_traveled decreases on a trip",
                    table,
                    indices,
                ]
            )

    # Check timepoint
    v = lambda x: x in range(2)
    problems = check_column(problems, table, f, "timepoint", v, column_required=False)

    if include_warnings:
        # Check for duplicated (trip_id, departure_time) pairs
        cond = f[["trip_id", "departure_time"]].duplicated()
        problems = check_table(
            problems,
            table,
            f,
            cond,
            "Repeated pair (trip_id, departure_time)",
            "warning",
        )

    return format_problems(problems, as_df=as_df)


def check_transfers(
    feed: "Feed", *, as_df: bool = False, include_warnings: bool = False
) -> list:
    """
    Analog of :func:`check_agency` for ``feed.transfers``.
    """
    table = "transfers"
    problems = []

    # Preliminary checks
    if feed.transfers is None:
        return problems

    f = feed.transfers.copy()
    problems = check_for_required_columns(problems, table, f)
    if problems:
        return format_problems(problems, as_df=as_df)

    if include_warnings:
        problems = check_for_invalid_columns(problems, table, f)

    # Check from_stop_id and to_stop_id
    for col in ["from_stop_id", "to_stop_id"]:
        problems = check_column_linked_id(
            problems, table, f, col, feed.stops, "stop_id"
        )

    # Check transfer_type
    v = lambda x: pd.isna(x) or x in range(5)
    problems = check_column(
        problems, table, f, "transfer_type", v, column_required=False
    )

    # Check min_transfer_time
    v = lambda x: x >= 0
    problems = check_column(
        problems, table, f, "min_transfer_time", v, column_required=False
    )

    return format_problems(problems, as_df=as_df)


def check_trips(
    feed: "Feed", *, as_df: bool = False, include_warnings: bool = False
) -> list:
    """
    Analog of :func:`check_agency` for ``feed.trips``.
    """
    table = "trips"
    problems = []

    # Preliminary checks
    if feed.trips is None:
        problems.append(["error", "Missing table", table, []])
    else:
        f = feed.trips.copy()
        problems = check_for_required_columns(problems, table, f)
    if problems:
        return format_problems(problems, as_df=as_df)

    if include_warnings:
        problems = check_for_invalid_columns(problems, table, f)

    # Check trip_id
    problems = check_column_id(problems, table, f, "trip_id")

    # Check route_id
    problems = check_column_linked_id(problems, table, f, "route_id", feed.routes)

    # Check service_id
    g = pd.DataFrame()
    if feed.calendar is not None:
        g = pd.concat([g, feed.calendar], sort=False)
    if feed.calendar_dates is not None:
        g = pd.concat([g, feed.calendar_dates], sort=False)
    problems = check_column_linked_id(problems, table, f, "service_id", g)

    # Check direction_id
    v = lambda x: x in range(2)
    problems = check_column(
        problems, table, f, "direction_id", v, column_required=False
    )

    # Check block_id
    if "block_id" in f.columns:
        v = lambda x: pd.isna(x) or valid_str(x)
        cond = ~f["block_id"].map(v)
        problems = check_table(problems, table, f, cond, "Blank block_id")

    # Check shape_id
    problems = check_column_linked_id(
        problems, table, f, "shape_id", feed.shapes, column_required=False
    )

    # Check wheelchair_accessible and bikes_allowed
    v = lambda x: x in range(3)
    for column in ["wheelchair_accessible", "bikes_allowed"]:
        problems = check_column(problems, table, f, column, v, column_required=False)

    # Check for trips with no stop times
    if include_warnings:
        s = feed.stop_times["trip_id"] if feed.stop_times is not None else []
        cond = ~f["trip_id"].isin(s)
        problems = check_table(
            problems, table, f, cond, "Trip has no stop times", "warning"
        )

    return format_problems(problems, as_df=as_df)


def validate(
    feed: "Feed", *, as_df: bool = True, include_warnings: bool = True
) -> Union[list, DataFrame]:
    """
    Check whether the given feed satisfies the GTFS.

    Parameters
    ----------
    feed : Feed
    as_df : boolean
        If ``True``, then return the resulting report as a DataFrame;
        otherwise return the result as a list
    include_warnings : boolean
        If ``True``, then include problems of types ``'error'`` and
        ``'warning'``; otherwise, only return problems of type
        ``'error'``

    Returns
    -------
    list or DataFrame
        Run all the table-checking functions: :func:`check_agency`,
        :func:`check_calendar`, etc.
        This yields a possibly empty list of items
        [problem type, message, table, rows].
        If ``as_df``, then format the error list as a DataFrame with the
        columns

        - ``'type'``: 'error' or 'warning'; 'error' means the GTFS is
          violated; 'warning' means there is a problem but it's not a
          GTFS violation
        - ``'message'``: description of the problem
        - ``'table'``: table in which problem occurs, e.g. 'routes'
        - ``'rows'``: rows of the table's DataFrame where problem occurs

        Return early if the feed is missing required tables or required
        columns.

    Notes
    -----
    - This function interprets the GTFS liberally, classifying problems
      as warnings rather than errors where the GTFS is unclear.
      For example if a trip_id listed in the trips table is not listed
      in the stop times table (a trip with no stop times),
      then that's a warning and not an error.
    - Timing benchmark: on a 2.80 GHz processor machine with 16 GB of
      memory, this function checks `this 31 MB Southeast Queensland feed
      <http://transitfeeds.com/p/translink/21/20170310>`_
      in 22 seconds, including warnings.

    """
    problems = []

    # Check for invalid columns and check the required tables
    checkers = [
        "check_agency",
        "check_calendar",
        "check_calendar_dates",
        "check_fare_attributes",
        "check_fare_rules",
        "check_feed_info",
        "check_frequencies",
        "check_routes",
        "check_shapes",
        "check_stops",
        "check_stop_times",
        "check_transfers",
        "check_trips",
    ]
    for checker in checkers:
        problems.extend(globals()[checker](feed, include_warnings=include_warnings))

    # Check calendar/calendar_dates combo
    if feed.calendar is None and feed.calendar_dates is None:
        problems.append(
            ["error", "Missing both tables", "calendar & calendar_dates", []]
        )

    return format_problems(problems, as_df=as_df)



"""
Functions about routes.
"""
from __future__ import annotations
from typing import Optional, Iterable, TYPE_CHECKING
import json

import geopandas as gp
import pandas as pd
import numpy as np
import shapely.geometry as sg
import shapely.ops as so
import folium as fl

from . import constants as cs
from . import helpers as hp

# Help mypy but avoid circular imports
if TYPE_CHECKING:
    from .feed import Feed


def compute_route_stats_0(
    trip_stats_subset: pd.DataFrame,
    headway_start_time: str = "07:00:00",
    headway_end_time: str = "19:00:00",
    *,
    split_directions: bool = False,
) -> pd.DataFrame:
    """
    Compute stats for the given subset of trips stats (of the form output by the
    function :func:`.trips.compute_trip_stats`).

    If ``split_directions``, then separate the stats by trip direction (0 or 1).
    Use the headway start and end times to specify the time period for computing
    headway stats.

    Return a DataFrame with the columns

    - ``'route_id'``
    - ``'route_short_name'``
    - ``'route_type'``
    - ``'direction_id'``
    - ``'num_trips'``: number of trips on the route in the subset
    - ``'num_trip_starts'``: number of trips on the route with
      nonnull start times
    - ``'num_trip_ends'``: number of trips on the route with nonnull
      end times that end before 23:59:59
    - ``'is_loop'``: 1 if at least one of the trips on the route has
      its ``is_loop`` field equal to 1; 0 otherwise
    - ``'is_bidirectional'``: 1 if the route has trips in both
      directions; 0 otherwise
    - ``'start_time'``: start time of the earliest trip on the route
    - ``'end_time'``: end time of latest trip on the route
    - ``'max_headway'``: maximum of the durations (in minutes)
      between trip starts on the route between
      ``headway_start_time`` and ``headway_end_time`` on the given
      dates
    - ``'min_headway'``: minimum of the durations (in minutes)
      mentioned above
    - ``'mean_headway'``: mean of the durations (in minutes)
      mentioned above
    - ``'peak_num_trips'``: maximum number of simultaneous trips in
      service (for the given direction, or for both directions when
      ``split_directions==False``)
    - ``'peak_start_time'``: start time of first longest period
      during which the peak number of trips occurs
    - ``'peak_end_time'``: end time of first longest period during
      which the peak number of trips occurs
    - ``'service_duration'``: total of the duration of each trip on
      the route in the given subset of trips; measured in hours
    - ``'service_distance'``: total of the distance traveled by each
      trip on the route in the given subset of trips;
      measured in kilometers if ``feed.dist_units`` is metric;
      otherwise measured in miles;
      contains all ``np.nan`` entries if ``feed.shapes is None``
    - ``'service_speed'``: service_distance/service_duration
    - ``'mean_trip_distance'``: service_distance/num_trips
    - ``'mean_trip_duration'``: service_duration/num_trips

    If not ``split_directions``, then remove the
    direction_id column and compute each route's stats,
    except for headways, using its trips running in both directions.
    In this case, (1) compute max headway by taking the max of the
    max headways in both directions; (2) compute mean headway by
    taking the weighted mean of the mean headways in both
    directions.

    If ``trip_stats_subset`` is empty, return an empty DataFrame.

    Raise a ValueError if ``split_directions`` and no non-NaN
    direction ID values present
    """
    if trip_stats_subset.empty:
        return pd.DataFrame()

    # Convert trip start and end times to seconds to ease calculations below
    f = trip_stats_subset.copy()
    f[["start_time", "end_time"]] = f[["start_time", "end_time"]].applymap(
        hp.timestr_to_seconds
    )

    headway_start = hp.timestr_to_seconds(headway_start_time)
    headway_end = hp.timestr_to_seconds(headway_end_time)

    def compute_route_stats_split_directions(group):
        # Take this group of all trips stats for a single route
        # and compute route-level stats.
        d = dict()
        d["route_short_name"] = group["route_short_name"].iat[0]
        d["route_type"] = group["route_type"].iat[0]
        d["num_trips"] = group.shape[0]
        d["num_trip_starts"] = group["start_time"].count()
        d["num_trip_ends"] = group.loc[
            group["end_time"] < 24 * 3600, "end_time"
        ].count()
        d["is_loop"] = int(group["is_loop"].any())
        d["start_time"] = group["start_time"].min()
        d["end_time"] = group["end_time"].max()

        # Compute max and mean headway
        stimes = group["start_time"].values
        stimes = sorted(
            [stime for stime in stimes if headway_start <= stime <= headway_end]
        )
        headways = np.diff(stimes)
        if headways.size:
            d["max_headway"] = np.max(headways) / 60  # minutes
            d["min_headway"] = np.min(headways) / 60  # minutes
            d["mean_headway"] = np.mean(headways) / 60  # minutes
        else:
            d["max_headway"] = np.nan
            d["min_headway"] = np.nan
            d["mean_headway"] = np.nan

        # Compute peak num trips
        active_trips = hp.get_active_trips_df(group[["start_time", "end_time"]])
        times, counts = active_trips.index.values, active_trips.values
        start, end = hp.get_peak_indices(times, counts)
        d["peak_num_trips"] = counts[start]
        d["peak_start_time"] = times[start]
        d["peak_end_time"] = times[end]

        d["service_distance"] = group["distance"].sum()
        d["service_duration"] = group["duration"].sum()

        return pd.Series(d)

    def compute_route_stats(group):
        d = dict()
        d["route_short_name"] = group["route_short_name"].iat[0]
        d["route_type"] = group["route_type"].iat[0]
        d["num_trips"] = group.shape[0]
        d["num_trip_starts"] = group["start_time"].count()
        d["num_trip_ends"] = group.loc[
            group["end_time"] < 24 * 3600, "end_time"
        ].count()
        d["is_loop"] = int(group["is_loop"].any())
        d["is_bidirectional"] = int(group["direction_id"].unique().size > 1)
        d["start_time"] = group["start_time"].min()
        d["end_time"] = group["end_time"].max()

        # Compute headway stats
        headways = np.array([])
        for direction in [0, 1]:
            stimes = group[group["direction_id"] == direction]["start_time"].values
            stimes = sorted(
                [stime for stime in stimes if headway_start <= stime <= headway_end]
            )
            headways = np.concatenate([headways, np.diff(stimes)])
        if headways.size:
            d["max_headway"] = np.max(headways) / 60  # minutes
            d["min_headway"] = np.min(headways) / 60  # minutes
            d["mean_headway"] = np.mean(headways) / 60  # minutes
        else:
            d["max_headway"] = np.nan
            d["min_headway"] = np.nan
            d["mean_headway"] = np.nan

        # Compute peak num trips
        active_trips = hp.get_active_trips_df(group[["start_time", "end_time"]])
        times, counts = active_trips.index.values, active_trips.values
        start, end = hp.get_peak_indices(times, counts)
        d["peak_num_trips"] = counts[start]
        d["peak_start_time"] = times[start]
        d["peak_end_time"] = times[end]

        d["service_distance"] = group["distance"].sum()
        d["service_duration"] = group["duration"].sum()

        return pd.Series(d)

    if split_directions:
        f = f.loc[lambda x: x.direction_id.notnull()].assign(
            direction_id=lambda x: x.direction_id.astype(int)
        )
        if f.empty:
            raise ValueError(
                "At least one trip stats direction ID value " "must be non-NaN."
            )

        g = (
            f.groupby(["route_id", "direction_id"])
            .apply(compute_route_stats_split_directions)
            .reset_index()
        )

        # Add the is_bidirectional column
        def is_bidirectional(group):
            d = {}
            d["is_bidirectional"] = int(group["direction_id"].unique().size > 1)
            return pd.Series(d)

        gg = g.groupby("route_id").apply(is_bidirectional).reset_index()
        g = g.merge(gg)
    else:
        g = f.groupby("route_id").apply(compute_route_stats).reset_index()

    # Compute a few more stats
    g["service_speed"] = (g["service_distance"] / g["service_duration"]).fillna(
        g["service_distance"]
    )
    g["mean_trip_distance"] = g["service_distance"] / g["num_trips"]
    g["mean_trip_duration"] = g["service_duration"] / g["num_trips"]

    # Convert route times to time strings
    g[["start_time", "end_time", "peak_start_time", "peak_end_time"]] = g[
        ["start_time", "end_time", "peak_start_time", "peak_end_time"]
    ].applymap(lambda x: hp.timestr_to_seconds(x, inverse=True))

    return g


def compute_route_time_series_0(
    trip_stats_subset: pd.DataFrame,
    date_label: str = "20010101",
    freq: str = "5Min",
    *,
    split_directions: bool = False,
) -> pd.DataFrame:
    """
    Compute stats in a 24-hour time series form for the given subset of trips (of the
    form output by the function :func:`.trips.compute_trip_stats`).

    If ``split_directions``, then separate each routes's stats by trip direction.
    Set the time series frequency according to the given frequency string;
    max frequency is one minute ('Min').
    Use the given YYYYMMDD date label as the date in the time series index.

    Return a DataFrame time series version the following route stats for each route.

    - ``num_trips``: number of trips in service on the route
      at any time within the time bin
    - ``num_trip_starts``: number of trips that start within
      the time bin
    - ``num_trip_ends``: number of trips that end within the
      time bin, ignoring trips that end past midnight
    - ``service_distance``: sum of the service distance accrued
      during the time bin across all trips on the route;
      measured in kilometers if ``feed.dist_units`` is metric;
      otherwise measured in miles;
    - ``service_duration``: sum of the service duration accrued
      during the time bin across all trips on the route;
      measured in hours
    - ``service_speed``: ``service_distance/service_duration``
      for the route

    The columns are hierarchical (multi-indexed) with

    - top level: name is ``'indicator'``; values are
      ``'num_trip_starts'``, ``'num_trip_ends'``, ``'num_trips'``,
      ``'service_distance'``, ``'service_duration'``, and
      ``'service_speed'``
    - middle level: name is ``'route_id'``;
      values are the active routes
    - bottom level: name is ``'direction_id'``; values are 0s and 1s

    If not ``split_directions``, then don't include the bottom level.

    The time series has a timestamp index for a 24-hour period
    sampled at the given frequency.
    The maximum allowable frequency is 1 minute.
    If ``trip_stats_subset`` is empty, then return an empty
    DataFrame with the columns ``'num_trip_starts'``,
    ``'num_trip_ends'``, ``'num_trips'``, ``'service_distance'``,
    ``'service_duration'``, and ``'service_speed'``.

    Notes
    -----
    - The time series is computed at a one-minute frequency, then
      resampled at the end to the given frequency
    - Trips that lack start or end times are ignored, so the the
      aggregate ``num_trips`` across the day could be less than the
      ``num_trips`` column of :func:`compute_route_stats_0`
    - All trip departure times are taken modulo 24 hours.
      So routes with trips that end past 23:59:59 will have all
      their stats wrap around to the early morning of the time series,
      except for their ``num_trip_ends`` indicator.
      Trip endings past 23:59:59 not binned so that resampling the
      ``num_trips`` indicator works efficiently.
    - Note that the total number of trips for two consecutive time bins
      t1 < t2 is the sum of the number of trips in bin t2 plus the
      number of trip endings in bin t1.
      Thus we can downsample the ``num_trips`` indicator by keeping
      track of only one extra count, ``num_trip_ends``, and can avoid
      recording individual trip IDs.
    - All other indicators are downsampled by summing.
    - Raise a ValueError if ``split_directions`` and no non-NaN
      direction ID values present

    """
    if trip_stats_subset.empty:
        return pd.DataFrame()

    tss = trip_stats_subset.copy()
    if split_directions:
        tss = tss.loc[lambda x: x.direction_id.notnull()].assign(
            direction_id=lambda x: x.direction_id.astype(int)
        )
        if tss.empty:
            raise ValueError(
                "At least one trip stats direction ID value " "must be non-NaN."
            )

        # Alter route IDs to encode direction:
        # <route ID>-0 and <route ID>-1 or <route ID>-NA
        tss["route_id"] = (
            tss["route_id"] + "-" + tss["direction_id"].map(lambda x: str(int(x)))
        )

    routes = tss["route_id"].unique()
    # Build a dictionary of time series and then merge them all
    # at the end.
    # Assign a uniform generic date for the index
    date_str = date_label
    day_start = pd.to_datetime(date_str + " 00:00:00")
    day_end = pd.to_datetime(date_str + " 23:59:00")
    rng = pd.period_range(day_start, day_end, freq="Min")
    indicators = [
        "num_trip_starts",
        "num_trip_ends",
        "num_trips",
        "service_duration",
        "service_distance",
    ]

    bins = [i for i in range(24 * 60)]  # One bin for each minute
    num_bins = len(bins)

    # Bin start and end times
    def F(x):
        return (hp.timestr_to_seconds(x) // 60) % (24 * 60)

    tss[["start_index", "end_index"]] = tss[["start_time", "end_time"]].applymap(F)
    routes = sorted(set(tss["route_id"].values))

    # Bin each trip according to its start and end time and weight
    series_by_route_by_indicator = {
        indicator: {route: [0 for i in range(num_bins)] for route in routes}
        for indicator in indicators
    }
    for index, row in tss.iterrows():
        route = row["route_id"]
        start = row["start_index"]
        end = row["end_index"]
        distance = row["distance"]

        if start is None or np.isnan(start) or start == end:
            continue

        # Get bins to fill
        if start <= end:
            bins_to_fill = bins[start:end]
        else:
            bins_to_fill = bins[start:] + bins[:end]

        # Bin trip.
        # Do num trip starts.
        series_by_route_by_indicator["num_trip_starts"][route][start] += 1
        # Don't mark trip ends for trips that run past midnight;
        # allows for easy resampling of num_trips later
        if start <= end:
            series_by_route_by_indicator["num_trip_ends"][route][end] += 1
        # Do rest of indicators
        for indicator in indicators[2:]:
            if indicator == "num_trips":
                weight = 1
            elif indicator == "service_duration":
                weight = 1 / 60
            else:
                weight = distance / len(bins_to_fill)
            for bin in bins_to_fill:
                series_by_route_by_indicator[indicator][route][bin] += weight

    # Create one time series per indicator
    rng = pd.date_range(date_str, periods=24 * 60, freq="Min")
    series_by_indicator = {
        indicator: pd.DataFrame(
            series_by_route_by_indicator[indicator], index=rng
        ).fillna(0)
        for indicator in indicators
    }

    # Combine all time series into one time series
    g = hp.combine_time_series(
        series_by_indicator, kind="route", split_directions=split_directions
    )

    return hp.downsample(g, freq=freq)


def get_routes(
    feed: "Feed", date: Optional[str] = None, time: Optional[str] = None
) -> pd.DataFrame:
    """
    Return ``feed.routes`` or a subset thereof.
    If a YYYYMMDD date string is given, then restrict routes to only those active on
    the date.
    If a HH:MM:SS time string is given, possibly with HH > 23, then restrict routes
    to only those active during the time.
    """
    if date is None:
        return feed.routes.copy()

    trips = feed.get_trips(date, time)
    R = trips["route_id"].unique()
    return feed.routes[feed.routes["route_id"].isin(R)]


def compute_route_stats(
    feed: "Feed",
    trip_stats_subset: pd.DataFrame,
    dates: list[str],
    headway_start_time: str = "07:00:00",
    headway_end_time: str = "19:00:00",
    *,
    split_directions: bool = False,
) -> pd.DataFrame:
    """
    Compute route stats for all the trips that lie in the given subset
    of trip stats (of the form output by the function :func:`.trips.compute_trip_stats`)
    and that start on the given dates (YYYYMMDD date strings).

    If ``split_directions``, then separate the stats by trip direction (0 or 1).
    Use the headway start and end times to specify the time period for computing
    headway stats.

    Return a DataFrame with the columns

    - ``'date'``
    - the columns listed in :func:``compute_route_stats_0``

    Exclude dates with no active trips, which could yield the empty DataFrame.

    Notes
    -----
    - The route stats for date d contain stats for trips that start on
      date d only and ignore trips that start on date d-1 and end on
      date d
    - Raise a ValueError if ``split_directions`` and no non-NaN
      direction ID values present

    """
    dates = feed.subset_dates(dates)
    if not dates:
        return pd.DataFrame()

    # Collect stats for each date,
    # memoizing stats the sequence of trip IDs active on the date
    # to avoid unnecessary recomputations.
    # Store in a dictionary of the form
    # trip ID sequence -> stats DataFarme.
    stats_by_ids = {}

    activity = feed.compute_trip_activity(dates)

    frames = []
    for date in dates:
        ids = tuple(activity.loc[activity[date] > 0, "trip_id"])
        if ids in stats_by_ids:
            stats = (
                stats_by_ids[ids]
                # Assign date
                .assign(date=date)
            )
        elif ids:
            # Compute stats
            t = trip_stats_subset.loc[lambda x: x.trip_id.isin(ids)].copy()
            stats = (
                compute_route_stats_0(
                    t,
                    split_directions=split_directions,
                    headway_start_time=headway_start_time,
                    headway_end_time=headway_end_time,
                )
                # Assign date
                .assign(date=date)
            )

            # Memoize stats
            stats_by_ids[ids] = stats
        else:
            stats = pd.DataFrame()

        frames.append(stats)

    # Assemble stats into a single DataFrame
    return pd.concat(frames)


def build_zero_route_time_series(
    feed: "Feed",
    date_label: str = "20010101",
    freq: str = "5Min",
    *,
    split_directions: bool = False,
) -> pd.DataFrame:
    """
    Return a route time series with the same index and hierarchical columns
    as output by :func:`compute_route_time_series_0`,
    but fill it full of zero values.
    """
    start = date_label
    end = pd.to_datetime(date_label + " 23:59:00")
    rng = pd.date_range(start, end, freq=freq)
    inds = [
        "num_trip_starts",
        "num_trip_ends",
        "num_trips",
        "service_duration",
        "service_distance",
        "service_speed",
    ]
    rids = feed.routes.route_id
    if split_directions:
        product = [inds, rids, [0, 1]]
        names = ["indicator", "route_id", "direction_id"]
    else:
        product = [inds, rids]
        names = ["indicator", "route_id"]
    cols = pd.MultiIndex.from_product(product, names=names)
    return pd.DataFrame([[0 for c in cols]], index=rng, columns=cols).sort_index(
        axis="columns"
    )


def compute_route_time_series(
    feed: "Feed",
    trip_stats_subset: pd.DataFrame,
    dates: list[str],
    freq: str = "5Min",
    *,
    split_directions: bool = False,
) -> pd.DataFrame:
    """
    Compute route stats in time series form for the trips that lie in
    the trip stats subset (of the form output by the function
    :func:`.trips.compute_trip_stats`) and that start on the given dates
    (YYYYMMDD date strings).

    If ``split_directions``, then separate each routes's stats by trip direction.
    Specify the time series frequency with a Pandas frequency string, e.g. ``'5Min'``;
    max frequency is one minute ('Min').

    Return a DataFrame of the same format output by the function
    :func:`compute_route_time_series_0` but with multiple dates

    Exclude dates that lie outside of the Feed's date range.
    If all dates lie outside the Feed's date range, then return an
    empty DataFrame.

    Notes
    -----
    - See the notes for :func:`compute_route_time_series_0`
    - Raise a ValueError if ``split_directions`` and no non-NaN
      direction ID values present

    """
    dates = feed.subset_dates(dates)
    if not dates:
        return pd.DataFrame()

    activity = feed.compute_trip_activity(dates)
    ts = trip_stats_subset.copy()

    # Collect stats for each date, memoizing stats by trip ID sequence
    # to avoid unnecessary re-computations.
    # Store in dictionary of the form
    # trip ID sequence ->
    # [stats DataFarme, date list that stats apply]
    stats_by_ids = {}
    zero_stats = build_zero_route_time_series(
        feed, split_directions=split_directions, freq=freq
    )
    for date in dates:
        ids = tuple(activity.loc[activity[date] > 0, "trip_id"])
        if ids in stats_by_ids:
            # Append date to date list
            stats_by_ids[ids][1].append(date)
        elif not ids:
            # Null stats
            stats_by_ids[ids] = [zero_stats, [date]]
        else:
            # Compute stats
            t = ts[ts["trip_id"].isin(ids)].copy()
            stats = compute_route_time_series_0(
                t, split_directions=split_directions, freq=freq, date_label=date
            )

            # Remember stats
            stats_by_ids[ids] = [stats, [date]]

    # Assemble stats into DataFrame
    frames = []
    for stats, dates_ in stats_by_ids.values():
        for date in dates_:
            f = stats.copy()
            # Replace date
            d = hp.datestr_to_date(date)
            f.index = f.index.map(
                lambda t: t.replace(year=d.year, month=d.month, day=d.day)
            )
            frames.append(f)

    f = pd.concat(frames).sort_index().sort_index(axis="columns")

    if len(dates) > 1:
        # Insert missing dates and zeros to complete series index
        end_datetime = pd.to_datetime(dates[-1] + " 23:59:59")
        new_index = pd.date_range(dates[0], end_datetime, freq=freq)
        f = f.reindex(new_index)
    else:
        # Set frequency
        f.index.freq = pd.tseries.frequencies.to_offset(freq)

    return f.rename_axis("datetime", axis="index")


def build_route_timetable(
    feed: "Feed", route_id: str, dates: list[str]
) -> pd.DataFrame:
    """
    Return a timetable for the given route and dates (YYYYMMDD date strings).

    Return a DataFrame with whose columns are all those in ``feed.trips`` plus those in
    ``feed.stop_times`` plus ``'date'``.
    The trip IDs are restricted to the given route ID.
    The result is sorted first by date and then by grouping by
    trip ID and sorting the groups by their first departure time.

    Skip dates outside of the Feed's dates.

    If there is no route activity on the given dates, then return
    an empty DataFrame.
    """
    dates = feed.subset_dates(dates)
    if not dates:
        return pd.DataFrame()

    t = pd.merge(feed.trips, feed.stop_times)
    t = t[t["route_id"] == route_id].copy()
    a = feed.compute_trip_activity(dates)

    frames = []
    for date in dates:
        # Slice to trips active on date
        ids = a.loc[a[date] == 1, "trip_id"]
        f = t[t["trip_id"].isin(ids)].copy()
        f["date"] = date
        # Groupby trip ID and sort groups by their minimum departure time.
        # For some reason NaN departure times mess up the transform below.
        # So temporarily fill NaN departure times as a workaround.
        f["dt"] = f["departure_time"].fillna(method="ffill")
        f["min_dt"] = f.groupby("trip_id")["dt"].transform(min)
        frames.append(f)

    f = pd.concat(frames)
    return f.sort_values(["date", "min_dt", "stop_sequence"]).drop(
        ["min_dt", "dt"], axis=1
    )


def geometrize_routes(
    feed: "Feed",
    route_ids: Optional[Iterable[str]] = None,
    *,
    use_utm: bool = False,
    split_directions: bool = False,
) -> gp.GeoDataFrame:
    """
    Given a Feed, return a GeoDataFrame with all the columns of ``feed.routes``
    plus a geometry column of (Multi)LineStrings, each of which represents the
    corresponding routes's shape.

    If an iterable of route IDs is given, then subset to those routes.
    If ``use_utm``, then use local UTM coordinates for the geometries.
    If ``split_directions``, then add the column ``direction_id`` and split each route
    route shapes into shapes in trip direction 0 and its shapes in trip direction 1.

    Raise a ValueError if the Feed has no shapes.
    """
    if feed.shapes is None:
        raise ValueError("This Feed has no shapes.")

    # Subset routes
    if route_ids is None:
        route_ids = feed.routes.route_id

    # Subset trips
    trip_ids = (
        feed.trips.loc[lambda x: x.route_id.isin(route_ids)]
        # Drop unnecessary duplicate shapes
        .drop_duplicates(subset="shape_id").loc[:, "trip_id"]
    )

    # Combine shape LineStrings within route and direction
    if split_directions:
        groupby_cols = ["route_id", "direction_id"]
    else:
        groupby_cols = ["route_id"]

    def merge_lines(group):
        d = {}
        d["geometry"] = so.linemerge(group.geometry.tolist())
        return pd.Series(d)

    if use_utm:
        lat, lon = feed.shapes[["shape_pt_lat", "shape_pt_lon"]].values[0]
        crs = hp.get_utm_crs(lat, lon)
    else:
        crs = cs.WGS84

    return (
        feed.geometrize_trips(trip_ids, use_utm=use_utm)
        .filter(["route_id", "direction_id", "geometry"])
        # GeoDataFrame disappears here
        .groupby(groupby_cols)
        .apply(merge_lines)
        .reset_index()
        .merge(feed.routes)
        .pipe(gp.GeoDataFrame, crs=crs)
    )


def routes_to_geojson(
    feed: "Feed",
    route_ids: Optional[Iterable[str]] = None,
    *,
    split_directions: bool = False,
    include_stops: bool = False,
) -> dict:
    """
    Return a GeoJSON FeatureCollection of MultiLineString features representing this Feed's routes.
    The coordinates reference system is the default one for GeoJSON,
    namely WGS84.

    If ``include_stops``, then include the route stops as Point features .
    If an iterable of route IDs is given, then subset to those routes.
    If the subset is empty, then return a FeatureCollection with an empty list of
    features.
    If the Feed has no shapes, then raise a ValueError.
    If any of the given route IDs are not found in the feed, then raise a ValueError.
    """
    if route_ids is None or not list(route_ids):
        route_ids = feed.routes.route_id

    D = set(route_ids) - set(feed.routes.route_id)
    if D:
        raise ValueError(f"Route IDs {D} not found in feed.")

    # Get routes
    g = geometrize_routes(feed, route_ids=route_ids, split_directions=split_directions)
    collection = json.loads(g.to_json())

    # Get stops if desired
    if include_stops:
        if route_ids is not None:
            stop_ids = (
                feed.stop_times.merge(feed.trips.filter(["trip_id", "route_id"]))
                .loc[lambda x: x.route_id.isin(route_ids), "stop_id"]
                .unique()
            )
        else:
            stop_ids = None

        stops_gj = feed.stops_to_geojson(stop_ids=stop_ids)
        collection["features"].extend(stops_gj["features"])

    return hp.drop_feature_ids(collection)


def map_routes(
    feed: "Feed",
    route_ids: Iterable[str],
    color_palette: list[str] = cs.COLORS_SET2,
    *,
    include_stops: bool = False,
):
    """
    Return a Folium map showing the given routes and (optionally)
    their stops.
    If any of the given route IDs are not found in the feed, then raise a ValueError.
    """
    # Initialize map
    my_map = fl.Map(tiles="cartodbpositron", prefer_canvas=True)

    # Create route colors
    n = len(route_ids)
    colors = [color_palette[i % len(color_palette)] for i in range(n)]

    # Collect route bounding boxes to set map zoom later
    bboxes = []

    # Create a feature group for each route and add it to the map
    for i, route_id in enumerate(route_ids):
        collection = feed.routes_to_geojson(
            route_ids=[route_id], include_stops=include_stops
        )

        # Use route short name for group name if possible; otherwise use route ID
        route_name = route_id
        for f in collection["features"]:
            if "route_short_name" in f["properties"]:
                route_name = f["properties"]["route_short_name"]
                break

        group = fl.FeatureGroup(name=f"Route {route_name}")
        color = colors[i]

        for f in collection["features"]:
            prop = f["properties"]

            # Add stop
            if f["geometry"]["type"] == "Point":
                lon, lat = f["geometry"]["coordinates"]
                fl.CircleMarker(
                    location=[lat, lon],
                    radius=8,
                    fill=True,
                    color=color,
                    weight=1,
                    popup=fl.Popup(hp.make_html(prop)),
                ).add_to(group)

            # Add path
            else:
                prop["color"] = color
                path = fl.GeoJson(
                    f,
                    name=prop["route_short_name"],
                    style_function=lambda x: {"color": x["properties"]["color"]},
                )
                path.add_child(fl.Popup(hp.make_html(prop)))
                path.add_to(group)
                bboxes.append(sg.box(*sg.shape(f["geometry"]).bounds))

        group.add_to(my_map)

    fl.LayerControl().add_to(my_map)

    # Fit map to bounds
    bounds = so.unary_union(bboxes).bounds
    bounds2 = [bounds[1::-1], bounds[3:1:-1]]  # Folium expects this ordering
    my_map.fit_bounds(bounds2)

    return my_map



"""
Functions useful across modules.
"""
from __future__ import annotations
import datetime as dt
from typing import Optional, Union, Callable
import copy
from bisect import bisect_left, bisect_right
from functools import cmp_to_key
import math

import pandas as pd
import numpy as np
import shapely.geometry as sg
from shapely.ops import transform
import utm
import json2html as j2h

from . import constants as cs


def datestr_to_date(
    x: Union[dt.date, str], format_str: str = "%Y%m%d", *, inverse: bool = False
) -> Union[str, dt.date]:
    """
    Given a string ``x`` representing a date in the given format,
    convert it to a datetime.date object and return the result.
    If ``inverse``, then assume that ``x`` is a date object and return
    its corresponding string in the given format.
    """
    if x is None:
        return None
    if not inverse:
        result = dt.datetime.strptime(x, format_str).date()
    else:
        result = x.strftime(format_str)
    return result


def timestr_to_seconds(
    x: Union[dt.date, str], *, inverse: bool = False, mod24: bool = False
) -> int:
    """
    Given an HH:MM:SS time string ``x``, return the number of seconds
    past midnight that it represents.
    In keeping with GTFS standards, the hours entry may be greater than
    23.
    If ``mod24``, then return the number of seconds modulo ``24*3600``.
    If ``inverse``, then do the inverse operation.
    In this case, if ``mod24`` also, then first take the number of
    seconds modulo ``24*3600``.
    """
    if not inverse:
        try:
            hours, mins, seconds = x.split(":")
            result = int(hours) * 3600 + int(mins) * 60 + int(seconds)
            if mod24:
                result %= 24 * 3600
        except:
            result = np.nan
    else:
        try:
            seconds = int(x)
            if mod24:
                seconds %= 24 * 3600
            hours, remainder = divmod(seconds, 3600)
            mins, secs = divmod(remainder, 60)
            result = f"{hours:02d}:{mins:02d}:{secs:02d}"
        except:
            result = np.nan
    return result


def timestr_mod24(timestr: str) -> int:
    """
    Given a GTFS HH:MM:SS time string, return a timestring in the same
    format but with the hours taken modulo 24.
    """
    try:
        hours, mins, secs = [int(x) for x in timestr.split(":")]
        hours %= 24
        result = f"{hours:02d}:{mins:02d}:{secs:02d}"
    except:
        result = None
    return result


def weekday_to_str(
    weekday: Union[int, str], *, inverse: bool = False
) -> Union[int, str]:
    """
    Given a weekday number (integer in the range 0, 1, ..., 6),
    return its corresponding weekday name as a lowercase string.
    Here 0 -> 'monday', 1 -> 'tuesday', and so on.
    If ``inverse``, then perform the inverse operation.
    """
    s = ["monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday"]
    if not inverse:
        try:
            return s[weekday]
        except:
            return
    else:
        try:
            return s.index(weekday)
        except:
            return


def get_segment_length(
    linestring: sg.LineString, p: sg.Point, q: Optional[sg.Point] = None
) -> float:
    """
    Given a Shapely linestring and two Shapely points,
    project the points onto the linestring, and return the distance
    along the linestring between the two points.
    If ``q is None``, then return the distance from the start of the
    linestring to the projection of ``p``.
    The distance is measured in the native coordinates of the linestring.
    """
    # Get projected distances
    d_p = linestring.project(p)
    if q is not None:
        d_q = linestring.project(q)
        d = abs(d_p - d_q)
    else:
        d = d_p
    return d


def get_max_runs(x) -> np.array:
    """
    Given a list of numbers, return a NumPy array of pairs
    (start index, end index + 1) of the runs of max value.

    Example::

        >>> get_max_runs([7, 1, 2, 7, 7, 1, 2])
        array([[0, 1],
               [3, 5]])

    Assume x is not empty.
    Recipe comes from
    `Stack Overflow <http://stackoverflow.com/questions/1066758/find-length-of-sequences-of-identical-values-in-a-numpy-array>`_.
    """
    # Get 0-1 array where 1 marks the max values of x
    x = np.array(x)
    m = np.max(x)
    y = (x == m) * 1
    # Bound y by zeros to detect runs properly
    bounded = np.hstack(([0], y, [0]))
    # Get 1 at run starts and -1 at run ends
    diffs = np.diff(bounded)
    run_starts = np.where(diffs > 0)[0]
    run_ends = np.where(diffs < 0)[0]
    return np.array([run_starts, run_ends]).T
    # # Get lengths of runs and find index of longest
    # idx = np.argmax(run_ends - run_starts)
    # return run_starts[idx], run_ends[idx]


def get_peak_indices(times: list, counts: list) -> np.array:
    """
    Given an increasing list of times as seconds past midnight and a
    list of trip counts at those respective times,
    return a pair of indices i, j such that times[i] to times[j] is
    the first longest time period such that for all i <= x < j,
    counts[x] is the max of counts.
    Assume times and counts have the same nonzero length.
    """
    max_runs = get_max_runs(counts)

    def get_duration(a):
        return times[a[1]] - times[a[0]]

    index = np.argmax(np.apply_along_axis(get_duration, 1, max_runs))
    return max_runs[index]


def is_metric(dist_units: str) -> bool:
    """
    Return True if the given distance units equals 'm' or 'km';
    otherwise return False.
    """
    return dist_units in ["m", "km"]


def get_convert_dist(
    dist_units_in: str, dist_units_out: str
) -> Callable[[float], float]:
    """
    Return a function of the form

      distance in the units ``dist_units_in`` ->
      distance in the units ``dist_units_out``

    Only supports distance units in :const:`constants.DIST_UNITS`.
    """
    di, do = dist_units_in, dist_units_out
    DU = cs.DIST_UNITS
    if not (di in DU and do in DU):
        raise ValueError(f"Distance units must lie in {DU}")

    d = {
        "ft": {"ft": 1, "m": 0.3048, "mi": 1 / 5280, "km": 0.000_304_8},
        "m": {"ft": 1 / 0.3048, "m": 1, "mi": 1 / 1609.344, "km": 1 / 1000},
        "mi": {"ft": 5280, "m": 1609.344, "mi": 1, "km": 1.609_344},
        "km": {"ft": 1 / 0.000_304_8, "m": 1000, "mi": 1 / 1.609_344, "km": 1},
    }
    return lambda x: d[di][do] * x


def almost_equal(f: pd.DataFrame, g: pd.DataFrame) -> bool:
    """
    Return ``True`` if and only if the given DataFrames are equal after
    sorting their columns names, sorting their values, and
    reseting their indices.
    """
    if f.empty or g.empty:
        return f.equals(g)
    else:
        # Put in canonical order
        F = f.sort_index(axis=1).sort_values(list(f.columns)).reset_index(drop=True)
        G = g.sort_index(axis=1).sort_values(list(g.columns)).reset_index(drop=True)
        return F.equals(G)


def is_not_null(df: pd.DataFrame, col_name: str) -> bool:
    """
    Return ``True`` if the given DataFrame has a column of the given
    name (string), and there exists at least one non-NaN value in that
    column; return ``False`` otherwise.
    """
    if (
        isinstance(df, pd.DataFrame)
        and col_name in df.columns
        and df[col_name].notnull().any()
    ):
        return True
    else:
        return False


def get_utm_crs(lat: float, lon: float) -> dict:
    """
    Return a GeoPandas coordinate reference system (CRS) string
    corresponding to the UTM projection appropriate to the given WGS84
    latitude and longitude.

    Code inspired by https://github.com/Turbo87/utm/issues/51.
    """
    zone = utm.from_latlon(lat, lon)[2]
    result = f"EPSG:326{zone:02d}" if lat >= 0 else f"EPSG:327{zone:02d}"
    return result


def linestring_to_utm(linestring: sg.LineString) -> sg.LineString:
    """
    Given a Shapely LineString in WGS84 coordinates,
    convert it to the appropriate UTM coordinates.
    If ``inverse``, then do the inverse.
    """

    def proj(x, y):
        return utm.from_latlon(y, x)[:2]

    return transform(proj, linestring)


def get_active_trips_df(trip_times: pd.DataFrame) -> pd.Series:
    """
    Count the number of trips in ``trip_times`` that are active
    at any given time.

    Assume ``trip_times`` contains the columns

    - start_time: start time of the trip in seconds past midnight
    - end_time: end time of the trip in seconds past midnight

    Return a Series whose index is times from midnight when trips
    start and end and whose values are the number of active trips for that time.
    """
    active_trips = (
        pd.concat(
            [
                pd.Series(1, trip_times.start_time),  # departed add 1
                pd.Series(-1, trip_times.end_time),  # arrived subtract 1
            ]
        )
        .groupby(level=0, sort=True)
        .sum()
        .cumsum()
        .ffill()
    )
    return active_trips


def combine_time_series(
    time_series_dict: dict[str, pd.DataFrame],
    kind: str,
    *,
    split_directions: bool = False,
) -> pd.DataFrame:
    """
    Combine the time series DataFrames in the given dictionary
    into one time series DataFrame with hierarchical columns.

    Parameters
    ----------
    time_series_dict : dictionary
        Has the form string -> time series
    kind : string
        ``'route'`` or ``'stop'``
    split_directions : boolean
        If ``True``, then assume the original time series contains data
        separated by trip direction; otherwise, assume not.
        The separation is indicated by a suffix ``'-0'`` (direction 0)
        or ``'-1'`` (direction 1) in the route ID or stop ID column
        values.

    Returns
    -------
    DataFrame
        Columns are hierarchical (multi-index).
        The top level columns are the keys of the dictionary and
        the second level columns are ``'route_id'`` and
        ``'direction_id'``, if ``kind == 'route'``, or 'stop_id' and
        ``'direction_id'``, if ``kind == 'stop'``.
        If ``split_directions``, then third column is
        ``'direction_id'``; otherwise, there is no ``'direction_id'``
        column.

    """
    if kind not in ["stop", "route"]:
        raise ValueError("kind must be 'stop' or 'route'")

    names = ["indicator"]
    if kind == "stop":
        names.append("stop_id")
    else:
        names.append("route_id")

    if split_directions:
        names.append("direction_id")

    def process_index(k):
        a, b = k.rsplit("-", 1)
        return a, int(b)

    frames = list(time_series_dict.values())
    new_frames = []
    if split_directions:
        for f in frames:
            ft = f.T
            ft.index = pd.MultiIndex.from_tuples(
                [process_index(k) for (k, __) in ft.iterrows()]
            )
            new_frames.append(ft.T)
    else:
        new_frames = frames
    result = pd.concat(
        new_frames, axis=1, keys=list(time_series_dict.keys()), names=names
    )

    return result.rename_axis("datetime", axis="index")


def downsample(time_series: pd.DataFrame, freq: str) -> pd.DataFrame:
    """
    Downsample the given route, stop, or feed time series,
    (outputs of :func:`.routes.compute_route_time_series`,
    :func:`.stops.compute_stop_time_series`, or
    :func:`.miscellany.compute_feed_time_series`,
    respectively) to the given Pandas frequency string (e.g. '15Min').
    Return the given time series unchanged if the given frequency is
    shorter than the original frequency.
    """

    f = time_series.copy()

    # Can't downsample to a shorter frequency
    if f.empty or pd.tseries.frequencies.to_offset(
        freq
    ) <= pd.tseries.frequencies.to_offset(pd.infer_freq(f.index)):
        return f

    result = None
    if "stop_id" in time_series.columns.names:
        # It's a stops time series
        result = f.resample(freq).sum(min_count=1)
    else:
        # It's a route or feed time series.
        inds = [
            "num_trips",
            "num_trip_starts",
            "num_trip_ends",
            "service_distance",
            "service_duration",
        ]
        frames = []

        # Resample num_trips in a custom way that depends on
        # num_trips and num_trip_ends
        def agg_num_trips(group):
            return group["num_trips"].iloc[-1] + group["num_trip_ends"].iloc[:-1].sum(
                min_count=1
            )

        num_trips = f.groupby(pd.Grouper(freq=freq)).apply(agg_num_trips)
        frames.append(num_trips)

        # Resample the rest of the indicators via summing, preserving all-NaNs
        frames.extend(
            [
                f[ind].resample(freq).agg(lambda x: x.sum(min_count=1))
                for ind in inds[1:]
            ]
        )

        g = pd.concat(frames, axis=1, keys=inds)

        # Calculate speed and add it to f. Can't resample it.
        speed = (g.service_distance / g.service_duration).fillna(g.service_distance)
        speed = pd.concat({"service_speed": speed}, axis=1)
        result = pd.concat([g, speed], axis=1)

    # Reset column names and sort the hierarchical columns to allow slicing;
    # see http://pandas.pydata.org/pandas-docs/stable/advanced.html#sorting-a-multiindex
    result.columns.names = f.columns.names
    result = result.sort_index(axis=1, sort_remaining=True)

    # Set frequency, which is not automatically set
    result.index.freq = freq

    return result


def unstack_time_series(time_series: pd.DataFrame) -> pd.DataFrame:
    """
    Given a route, stop, or feed time series of the form output by the functions,
    :func:`compute_stop_time_series`, :func:`compute_route_time_series`, or
    :func:`compute_feed_time_series`, respectively, unstack it to return a DataFrame
    of with the columns:

    - ``"datetime"``
    - the columns ``time_series.columns.names``
    - ``"value"``: value at the datetime and other columns

    """
    col_names = time_series.columns.names
    return (
        time_series.unstack()
        .pipe(pd.DataFrame)
        .reset_index()
        .rename(columns={0: "value", "level_2": "datetime"})
        # Reorder columns
        .filter(["datetime"] + col_names + ["value"])
        .sort_values(["datetime"] + col_names)
    )


def restack_time_series(unstacked_time_series: pd.DataFrame) -> pd.DataFrame:
    """
    Given an unstacked stop, route, or feed time series in the form
    output by the function :func:`unstack_time_series`, restack it into
    its original time series form.
    """
    f = unstacked_time_series
    columns = [c for c in f.columns if c not in ["datetime", "value"]]
    g = f.pivot_table(index="datetime", columns=columns).value.sort_index(
        axis="columns"
    )

    # Get time series frequency
    if g.index.size > 1:
        hours = (g.index[1] - g.index[0]).components.hours
        if hours != 0:
            freq = f"{hours}H"
        else:
            freq = "D"
    else:
        freq = "D"

    # If necessary, insert missing dates and NaNs to complete series index
    num_dates = len(set(g.index.date))
    if num_dates > 1:
        end_datetime = pd.to_datetime(f"{g.index.date[-1]:%Y-%m-%d}" + " 23:59:59")
        new_index = pd.date_range(g.index[0], end_datetime, freq=freq, name="datetime")
        g = g.reindex(new_index)

    g.index.freq = freq

    return g


def make_html(d: dict) -> str:
    """
    Convert the given dictionary into an HTML table (string) with
    two columns: keys of dictionary, values of dictionary.
    """
    return j2h.json2html.convert(
        json=d, table_attributes="class='table table-condensed table-hover'"
    )


def drop_feature_ids(collection: dict) -> dict:
    """
    Given a GeoJSON FeatureCollection, remove the ``'id'`` attribute of each
    Feature, if it exists.
    """
    new_features = []
    for f in collection["features"]:
        new_f = copy.deepcopy(f)
        if "id" in new_f:
            del new_f["id"]
        new_features.append(new_f)

    collection["features"] = new_features
    return collection


def longest_subsequence(
    seq, mode="strictly", order="increasing", key=None, *, index=False
):
    """
    Return the longest increasing subsequence of `seq`.

    Parameters
    ----------
    seq : sequence object
      Can be any sequence, like `str`, `list`, `numpy.array`.
    mode : {'strict', 'strictly', 'weak', 'weakly'}, optional
      If set to 'strict', the subsequence will contain unique elements.
      Using 'weak' an element can be repeated many times.
      Modes ending in -ly serve as a convenience to use with `order` parameter,
      because `longest_sequence(seq, 'weakly', 'increasing')` reads better.
      The default is 'strict'.
    order : {'increasing', 'decreasing'}, optional
      By default return the longest increasing subsequence, but it is possible
      to return the longest decreasing sequence as well.
    key : function, optional
      Specifies a function of one argument that is used to extract a comparison
      key from each list element (e.g., `str.lower`, `lambda x: x[0]`).
      The default value is `None` (compare the elements directly).
    index : bool, optional
      If set to `True`, return the indices of the subsequence, otherwise return
      the elements. Default is `False`.

    Returns
    -------
    elements : list, optional
      A `list` of elements of the longest subsequence.
      Returned by default and when `index` is set to `False`.
    indices : list, optional
      A `list` of indices pointing to elements in the longest subsequence.
      Returned when `index` is set to `True`.

    Taken from `this Stack Overflow answer <https://stackoverflow.com/a/38337443>`_.
    """
    bisect = bisect_left if mode.startswith("strict") else bisect_right

    # compute keys for comparison just once
    rank = seq if key is None else map(key, seq)
    if order == "decreasing":
        rank = map(cmp_to_key(lambda x, y: 1 if x < y else 0 if x == y else -1), rank)
    rank = list(rank)

    if not rank:
        return []

    lastoflength = [0]  # end position of subsequence with given length
    predecessor = [None]  # penultimate element of l.i.s. ending at given position

    for i in range(1, len(seq)):
        # seq[i] can extend a subsequence that ends with a lesser (or equal) element
        j = bisect([rank[k] for k in lastoflength], rank[i])
        # update existing subsequence of length j or extend the longest
        try:
            lastoflength[j] = i
        except:
            lastoflength.append(i)
        # remember element before seq[i] in the subsequence
        predecessor.append(lastoflength[j - 1] if j > 0 else None)

    # trace indices [p^n(i), ..., p(p(i)), p(i), i], where n=len(lastoflength)-1
    def trace(i):
        if i is not None:
            yield from trace(predecessor[i])
            yield i

    indices = trace(lastoflength[-1])

    return list(indices) if index else [seq[i] for i in indices]


def make_ids(n: int, prefix: str = "id_"):
    """
    Return a length ``n`` list of unique sequentially labelled strings for use as IDs.

    Example::

        >>> make_ids(11, prefix="s")
        ['s00', s01', 's02', 's03', 's04', 's05', 's06', 's07', 's08', 's09', 's10']

    """
    if n < 1:
        result = []
    elif n == 1:
        result = [f"{prefix}0"]
    else:
        k = int(math.log10(n - 1)) + 1  # Number of digits for IDs
        result = [f"{prefix}{i:0{k}d}" for i in range(n)]

    return result



"""
This module defines a Feed class to represent GTFS feeds.
There is an instance attribute for every GTFS table (routes, stops, etc.),
which stores the table as a Pandas DataFrame,
or as ``None`` in case that table is missing.

The Feed class also has heaps of methods: a method to compute route stats,
a method to compute screen line counts, validations methods, etc.
To ease reading, almost all of these methods are defined in other modules and
grouped by theme (``routes.py``, ``stops.py``, etc.).
These methods, or rather functions that operate on feeds, are
then imported within the Feed class.
This separation of methods unfortunately messes up slightly the ``Feed`` class
documentation generated by Sphinx, introducing an extra leading ``feed``
parameter in the method signatures.
Ignore that extra parameter; it refers to the Feed instance,
usually called ``self`` and usually hidden automatically by Sphinx.
"""
from pathlib import Path
import tempfile
import shutil
from copy import deepcopy
import zipfile
from typing import Optional, Union

import pandas as pd
from pandas.core.frame import DataFrame
import requests

from . import constants as cs
from . import helpers as hp
from . import cleaners as cn


class Feed(object):
    """
    An instance of this class represents a not-necessarily-valid GTFS feed,
    where GTFS tables are stored as DataFrames.
    Beware that the stop times DataFrame can be big (several gigabytes),
    so make sure you have enough memory to handle it.

    Primary instance attributes:

    - ``dist_units``: a string in :const:`.constants.DIST_UNITS`;
      specifies the distance units of the `shape_dist_traveled` column values,
      if present; also effects whether to display trip and route stats in
      metric or imperial units
    - ``agency``
    - ``stops``
    - ``routes``
    - ``trips``
    - ``stop_times``
    - ``calendar``
    - ``calendar_dates``
    - ``fare_attributes``
    - ``fare_rules``
    - ``shapes``
    - ``frequencies``
    - ``transfers``
    - ``feed_info``

    There are also a few secondary instance attributes that are derived
    from the primary attributes and are automatically updated when the
    primary attributes change.
    However, for this update to work, you must update the primary
    attributes like this (good)::

        feed.trips['route_short_name'] = 'bingo'
        feed.trips = feed.trips

    and **not** like this (bad)::

        feed.trips['route_short_name'] = 'bingo'

    The first way ensures that the altered trips DataFrame is saved as
    the new ``trips`` attribute, but the second way does not.

    """

    # Import heaps of methods from modules split by functionality;
    # i learned this trick from
    # https://groups.google.com/d/msg/comp.lang.python/goLBrqcozNY/DPgyaZ6gAwAJ
    from .calendar import get_dates, get_week, get_first_week, subset_dates
    from .routes import (
        get_routes,
        compute_route_stats,
        build_zero_route_time_series,
        compute_route_time_series,
        build_route_timetable,
        geometrize_routes,
        routes_to_geojson,
        map_routes,
    )
    from .shapes import (
        append_dist_to_shapes,
        geometrize_shapes,
        build_geometry_by_shape,
        shapes_to_geojson,
        get_shapes_intersecting_geometry,
    )
    from .stops import (
        get_stops,
        compute_stop_activity,
        compute_stop_stats,
        build_zero_stop_time_series,
        compute_stop_time_series,
        build_stop_timetable,
        geometrize_stops,
        build_geometry_by_stop,
        stops_to_geojson,
        get_stops_in_area,
        map_stops,
    )
    from .stop_times import (
        get_stop_times,
        append_dist_to_stop_times,
        get_start_and_end_times,
        stop_times_to_geojson,
    )
    from .trips import (
        is_active_trip,
        get_trips,
        compute_trip_activity,
        compute_busiest_date,
        compute_trip_stats,
        locate_trips,
        geometrize_trips,
        trips_to_geojson,
        map_trips,
    )
    from .miscellany import (
        summarize,
        describe,
        assess_quality,
        convert_dist,
        compute_feed_stats,
        compute_feed_time_series,
        create_shapes,
        compute_bounds,
        compute_convex_hull,
        compute_centroid,
        restrict_to_dates,
        restrict_to_routes,
        restrict_to_area,
        compute_screen_line_counts,
    )
    from .validators import (
        validate,
        check_agency,
        check_calendar,
        check_calendar_dates,
        check_fare_attributes,
        check_fare_rules,
        check_feed_info,
        check_frequencies,
        check_routes,
        check_shapes,
        check_stops,
        check_stop_times,
        check_transfers,
        check_trips,
    )
    from .cleaners import (
        clean_ids,
        clean_times,
        clean_route_short_names,
        drop_zombies,
        aggregate_routes,
        aggregate_stops,
        clean,
        drop_invalid_columns,
    )

    def __init__(
        self,
        dist_units: str,
        agency: Optional[DataFrame] = None,
        stops: Optional[DataFrame] = None,
        routes: Optional[DataFrame] = None,
        trips: Optional[DataFrame] = None,
        stop_times: Optional[DataFrame] = None,
        calendar: Optional[DataFrame] = None,
        calendar_dates: Optional[DataFrame] = None,
        fare_attributes: Optional[DataFrame] = None,
        fare_rules: Optional[DataFrame] = None,
        shapes: Optional[DataFrame] = None,
        frequencies: Optional[DataFrame] = None,
        transfers: Optional[DataFrame] = None,
        feed_info: Optional[DataFrame] = None,
    ):
        """
        Assume that every non-None input is a DataFrame,
        except for ``dist_units`` which should be a string in
        :const:`.constants.DIST_UNITS`.

        No other format checking is performed.
        In particular, a Feed instance need not represent a valid GTFS
        feed.
        """
        # Set primary attributes from inputs.
        # The @property magic below will then
        # validate some and set some derived attributes
        for prop, val in locals().items():
            if prop in cs.FEED_ATTRS_1:
                setattr(self, prop, val)

    @property
    def dist_units(self):
        """
        The distance units of the Feed.
        """
        return self._dist_units

    @dist_units.setter
    def dist_units(self, val):
        if val not in cs.DIST_UNITS:
            raise ValueError(
                f"Distance units are required and " f"must lie in {cs.DIST_UNITS}"
            )
        else:
            self._dist_units = val

    @property
    def trips(self):
        """
        The trips table of this Feed.
        """
        return self._trips

    @trips.setter
    def trips(self, val):
        """
        Update ``self._trips_i`` if ``self.trips`` changes.
        """
        self._trips = val
        if val is not None and not val.empty:
            self._trips_i = self._trips.set_index("trip_id")
        else:
            self._trips_i = None

    @property
    def calendar(self):
        """
        The calendar table of this Feed.
        """
        return self._calendar

    @calendar.setter
    def calendar(self, val):
        """
        Update ``self._calendar_i``if ``self.calendar`` changes.
        """
        self._calendar = val
        if val is not None and not val.empty:
            self._calendar_i = self._calendar.set_index("service_id")
        else:
            self._calendar_i = None

    @property
    def calendar_dates(self):
        """
        The calendar_dates table of this Feed.
        """
        return self._calendar_dates

    @calendar_dates.setter
    def calendar_dates(self, val):
        """
        Update ``self._calendar_dates_g``
        if ``self.calendar_dates`` changes.
        """
        self._calendar_dates = val
        if val is not None and not val.empty:
            self._calendar_dates_g = self._calendar_dates.groupby(
                ["service_id", "date"]
            )
        else:
            self._calendar_dates_g = None

    def __str__(self):
        """
        Print the first five rows of each GTFS table.
        """
        d = {}
        for table in cs.GTFS_REF["table"].unique():
            try:
                d[table] = getattr(self, table).head(5)
            except:
                d[table] = None
        d["dist_units"] = self.dist_units

        return "\n".join([f"* {k} --------------------\n\t{v}" for k, v in d.items()])

    def __eq__(self, other):
        """
        Define two feeds be equal if and only if their
        :const:`.constants.FEED_ATTRS` attributes are equal,
        or almost equal in the case of DataFrames
        (but not groupby DataFrames).
        Almost equality is checked via :func:`.helpers.almost_equal`,
        which   canonically sorts DataFrame rows and columns.
        """
        # Return False if failures
        for key in cs.FEED_ATTRS_1:
            x = getattr(self, key)
            y = getattr(other, key)
            # DataFrame case
            if isinstance(x, pd.DataFrame):
                if not isinstance(y, pd.DataFrame) or not hp.almost_equal(x, y):
                    return False
            # Other case
            else:
                if x != y:
                    return False
        # No failures
        return True

    def copy(self) -> "Feed":
        """
        Return a copy of this feed, that is, a feed with all the same
        attributes.
        """
        other = Feed(dist_units=self.dist_units)
        for key in set(cs.FEED_ATTRS) - set(["dist_units"]):
            value = getattr(self, key)
            if isinstance(value, pd.DataFrame):
                # Pandas copy DataFrame
                value = value.copy()
            elif isinstance(value, pd.core.groupby.DataFrameGroupBy):
                # Pandas does not have a copy method for groupby objects
                # as far as i know
                value = deepcopy(value)
            setattr(other, key, value)

        return other

    def write(self, path: Path, ndigits: int = 6) -> None:
        """
        Write this Feed to the given path.
        If the path end in '.zip', then write the feed as a zip archive.
        Otherwise assume the path is a directory, and write the feed as a
        collection of CSV files to that directory, creating the directory
        if it does not exist.
        Round all decimals to ``ndigits`` decimal places.
        All distances will be the distance units ``feed.dist_units``.
        """
        path = Path(path)

        if path.suffix == ".zip":
            # Write to temporary directory before zipping
            zipped = True
            tmp_dir = tempfile.TemporaryDirectory()
            new_path = Path(tmp_dir.name)
        else:
            zipped = False
            if not path.exists():
                path.mkdir()
            new_path = path

        for table in cs.GTFS_REF["table"].unique():
            f = getattr(self, table)
            if f is None:
                continue

            f = f.copy()
            # Some columns need to be output as integers.
            # If there are NaNs in any such column,
            # then Pandas will format the column as float, which we don't want.
            f_int_cols = set(cs.INT_COLS) & set(f.columns)
            for s in f_int_cols:
                f[s] = f[s].fillna(-1).astype(int).astype(str).replace("-1", "")
            p = new_path / (table + ".txt")
            f.to_csv(str(p), index=False, float_format=f"%.{ndigits}f")

        # Zip directory
        if zipped:
            basename = str(path.parent / path.stem)
            shutil.make_archive(basename, format="zip", root_dir=tmp_dir.name)
            tmp_dir.cleanup()


# -------------------------------------
# Functions about input and output
# -------------------------------------
def list_feed(path: Path) -> DataFrame:
    """
    Given a path (string or Path object) to a GTFS zip file or
    directory, record the file names and file sizes of the contents,
    and return the result in a DataFrame with the columns:

    - ``'file_name'``
    - ``'file_size'``
    """
    path = Path(path)
    if not path.exists():
        raise ValueError(f"Path {path} does not exist")

    # Collect rows of DataFrame
    rows = []
    if path.is_file():
        # Zip file
        with zipfile.ZipFile(str(path)) as src:
            for x in src.infolist():
                if x.filename == "./":
                    continue
                d = {}
                d["file_name"] = x.filename
                d["file_size"] = x.file_size
                rows.append(d)
    else:
        # Directory
        for x in path.iterdir():
            d = {}
            d["file_name"] = x.name
            d["file_size"] = x.stat().st_size
            rows.append(d)

    return pd.DataFrame(rows)


def _read_feed_from_path(path: Path, dist_units: str) -> "Feed":
    """
    Helper function for :func:`read_feed`.
    Create a Feed instance from the given path and given distance units.
    The path should be a directory containing GTFS text files or a
    zip file that unzips as a collection of GTFS text files
    (and not as a directory containing GTFS text files).
    The distance units given must lie in :const:`constants.dist_units`

    Notes:

    - Ignore non-GTFS files in the feed
    - Automatically strip whitespace from the column names in GTFS files

    """
    path = Path(path)
    if not path.exists():
        raise ValueError(f"Path {path} does not exist")

    # Unzip path to temporary directory if necessary
    if path.is_file():
        zipped = True
        tmp_dir = tempfile.TemporaryDirectory()
        src_path = Path(tmp_dir.name)
        shutil.unpack_archive(str(path), tmp_dir.name, "zip")
    else:
        zipped = False
        src_path = path

    # Read files into feed dictionary of DataFrames
    feed_dict = {table: None for table in cs.GTFS_REF["table"]}
    for p in src_path.iterdir():
        table = p.stem
        # Skip empty files, irrelevant files, and files with no data
        if (
            p.is_file()
            and p.stat().st_size
            and p.suffix == ".txt"
            and table in feed_dict
        ):
            # utf-8-sig gets rid of the byte order mark (BOM);
            # see http://stackoverflow.com/questions/17912307/u-ufeff-in-python-string
            df = pd.read_csv(p, dtype=cs.DTYPE, encoding="utf-8-sig")
            if not df.empty:
                feed_dict[table] = cn.clean_column_names(df)

    feed_dict["dist_units"] = dist_units

    # Delete temporary directory
    if zipped:
        tmp_dir.cleanup()

    # Create feed
    return Feed(**feed_dict)


def _read_feed_from_url(url: str, dist_units: str) -> "Feed":
    """
    Helper function for :func:`read_feed`.
    Create a Feed instance from the given URL and given distance units.
    Assume the URL is valid and let the Requests library raise any errors.

    Notes:

    - Ignore non-GTFS files in the feed
    - Automatically strip whitespace from the column names in GTFS files


    """
    f = tempfile.NamedTemporaryFile(delete=False)
    with requests.get(url) as r:
        f.write(r._content)
    f.close()
    feed = _read_feed_from_path(f.name, dist_units=dist_units)
    Path(f.name).unlink()
    return feed


def read_feed(path_or_url: Union[Path, str], dist_units: str) -> "Feed":
    """
    Create a Feed instance from the given path or URL and given distance units.
    If the path exists, then call :func:`_read_feed_from_path`.
    Else if the URL has OK status according to Requests, then call
    :func:`_read_feed_from_url`.
    Else raise a ValueError.

    Notes:

    - Ignore non-GTFS files in the feed
    - Automatically strip whitespace from the column names in GTFS files

    """
    try:
        path_exists = Path(path_or_url).exists()
    except OSError:
        path_exists = False
    if path_exists:
        return _read_feed_from_path(path_or_url, dist_units=dist_units)
    elif requests.head(path_or_url).ok:
        return _read_feed_from_url(path_or_url, dist_units=dist_units)
    else:
        raise ValueError("Path does not exist or URL has bad status.")



"""
Functions about calendar and calendar_dates.
"""
from __future__ import annotations
import dateutil.relativedelta as rd
from typing import TYPE_CHECKING

from . import helpers as hp

# Help mypy but avoid circular imports
if TYPE_CHECKING:
    from .feed import Feed


def get_dates(feed: "Feed", *, as_date_obj: bool = False) -> list[str]:
    """
    Return a list of YYYYMMDD date strings for which the given Feed is valid,
    which could be the empty list if the Feed has no calendar information.

    If ``as_date_obj``, then return datetime.date objects instead.
    """
    dates = []
    if feed.calendar is not None and not feed.calendar.empty:
        if "start_date" in feed.calendar.columns:
            dates.append(feed.calendar["start_date"].min())
        if "end_date" in feed.calendar.columns:
            dates.append(feed.calendar["end_date"].max())
    if feed.calendar_dates is not None and not feed.calendar_dates.empty:
        if "date" in feed.calendar_dates.columns:
            start = feed.calendar_dates["date"].min()
            end = feed.calendar_dates["date"].max()
            dates.extend([start, end])
    if not dates:
        return []

    start_date, end_date = min(dates), max(dates)
    start_date, end_date = map(hp.datestr_to_date, [start_date, end_date])
    num_days = (end_date - start_date).days
    result = [start_date + rd.relativedelta(days=+d) for d in range(num_days + 1)]

    # Convert dates back to strings if required
    if not as_date_obj:
        result = [hp.datestr_to_date(x, inverse=True) for x in result]

    return result


def get_week(feed: "Feed", k: int, *, as_date_obj: bool = False) -> list[str]:
    """
    Given a Feed and a positive integer ``k``,
    return a list of YYYYMMDD date strings corresponding to the kth Monday--Sunday week
    (or initial segment thereof) for which the Feed is valid.
    For example, k=1 returns the first Monday--Sunday week (or initial segment thereof).
    If the Feed does not have k Mondays, then return the empty list.

    If ``as_date_obj``, then return datetime.date objects instead.
    """
    dates = feed.get_dates(as_date_obj=True)
    n = len(dates)

    # Get first Monday
    monday_index = None
    for (i, date) in enumerate(dates):
        if date.weekday() == 0:
            monday_index = i
            break

    # Get week k
    if k < 1 or monday_index is None or monday_index + 7 * (k - 1) > n:
        result = []
    else:
        result = dates[monday_index + 7 * (k - 1) : monday_index + 7 * k]

    # Convert to date strings if requested
    if not as_date_obj:
        result = [hp.datestr_to_date(x, inverse=True) for x in result]

    return result


def get_first_week(feed: "Feed", *, as_date_obj: bool = False) -> list[str]:
    """
    Return a list of YYYYMMDD date strings for the first Monday--Sunday
    week (or initial segment thereof) for which the given Feed is valid.
    If the feed has no Mondays, then return the empty list.

    If ``as_date_obj``, then return date objects, otherwise return date strings.
    """
    return get_week(feed, 1, as_date_obj=as_date_obj)


def subset_dates(feed: "Feed", dates: list[str]) -> list[str]:
    """
    Given a Feed and a list of YYYYMMDD date strings,
    return the sublist of dates that lie in the Feed's dates
    (the output :func:`feed.get_dates`).
    """
    return [d for d in dates if d in feed.get_dates()]
